from __future__ import annotations

"""
SQLite Backend

Borrowed parts from `python-diskache <https://github.com/grantjenks/python-diskcache>`_
"""

import abc
import copy
import time
import asyncio
import contextlib
import threading
import contextlib
import typing as t
import atexit
from lzl import load
from lzl.pool import ThreadPool
from lzo.types import eproperty
from pydantic.networks import UrlConstraints, Annotated
from pydantic.networks import UrlConstraints, Annotated
try:
    from pydantic_core import MultiHostUrl
except ImportError:
    from pydantic.networks import MultiHostUrl

from pydantic.alias_generators import to_camel
from ..base import logger
from .utils import dict_diff
from typing import TypeVar, Generic, Any, Dict, Optional, Union, Tuple, Iterable, List, Type, Callable, Generator, AsyncGenerator, TYPE_CHECKING

try:
    import dill as pickle
except ImportError:
    import pickle

if load.TYPE_CHECKING:
    import sqlite3
    import aiosqlite

    from sqlite3 import Connection, Cursor
    from aiosqlite import Connection as AsyncConnection, Cursor as AsyncCursor
else:
    sqlite3 = load.LazyLoad("sqlite3")
    aiosqlite = load.LazyLoad("aiosqlite")

class Constant(tuple):
    """Pretty display of immutable constant."""

    def __new__(cls, name):
        return tuple.__new__(cls, (name,))

    def __repr__(self):
        return f'{self[0]}'


ENOVAL = Constant('ENOVAL')
UNKNOWN = Constant('UNKNOWN')

class Timeout(Exception):
    """Database timeout expired."""

SqliteDsn = Annotated[
    MultiHostUrl,
    UrlConstraints(
        host_required=True,
        allowed_schemes=[
            'sqlite',
            'sqlite+aiosqlite',
            'sqlite+asyncio',
            'sqlite+aiosqlite3',
            'sqlite+pysqlite3',

            # Support rqlite
        ],
    ),
]


DEFAULT_SETTINGS = {
    'statistics': 1,  # True
    'tag_index': 1,  # True
    'eviction_policy': 'least-recently-stored',
    'size_limit': 2**30,  # 1gb
    'cull_limit': 10,
    'sqlite_auto_vacuum': 1,  # FULL
    'sqlite_cache_size': 2**13,  # 8,192 pages
    'sqlite_journal_mode': 'wal',
    'sqlite_mmap_size': 2**26,  # 64mb
    'sqlite_synchronous': 1,  # NORMAL
}

OPTIMIZED = {
    'standard': {
        'statistics': 0,
        'tag_index': 0,
    },
    'balanced': {
        'sqlite_mmap_size': 1024 * 128,  # 128mb
    },
    'cache': {
        'eviction_policy': 'least-frequently-used',
        'sqlite_mmap_size': 1024 * 512,  # 512mb
        'size_limit': 1024 * 1024 * 1024 * 5,  # 5gb
    },
    'write': {
        'sqlite_mmap_size': 1024 * 256,  # 256mb
        'size_limit': 1024 * 1024 * 1024 * 2,  # 2gb
    },
    'read': {
        'sqlite_mmap_size': 1024 * 128,  # 128mb
        'size_limit': 1024 * 1024 * 1024 * 1,  # 1gb
    },
}

METADATA = {
    'count': 0,
    'size': 0,
    'hits': 0,
    'misses': 0,
}

Optimization = t.Literal[
    'standard', 
    'balanced', 
    'cache',
    'write',
    'read',
]

def get_optimized_settings(
    optimization: Optional[str] = None, 
    **kwargs
) -> Dict[str, Any]:
    """
    Returns the Optimized Settings
    """
    base = copy.deepcopy(DEFAULT_SETTINGS)
    if optimization:
        base.update(OPTIMIZED.get(optimization, {}))
    if kwargs: base.update(kwargs)
    return base

def get_eviction_policies(table_name: str) -> Dict[str, t.Optional[str]]:
    """
    Returns the Eviction Policies
    """

    return {
        'none': {
            'init': None,
            'get': 'hit_count = hit_count + 1',
            'cull': None,
        },
        'least-recently-stored': {
            'init': (
                f'CREATE INDEX IF NOT EXISTS "{table_name}_store_time" ON'
                f' "{table_name}" (store_time)'
            ),
            # 'get': None,
            'get': 'hit_count = hit_count + 1',
            'cull': 'SELECT {fields} FROM "' + table_name + '" ORDER BY store_time LIMIT ?',
        },
        'least-recently-used': {
            'init': (
                f'CREATE INDEX IF NOT EXISTS "{table_name}_access_time" ON'
                f' "{table_name}" (access_time)'
            ),
            'get': 'access_time = {now}, hit_count = hit_count + 1',
            'cull': 'SELECT {fields} FROM "' + table_name + '" ORDER BY access_time LIMIT ?',
        },
        'least-frequently-used': {
            'init': (
                f'CREATE INDEX IF NOT EXISTS "{table_name}_access_count" ON'
                f' "{table_name}" (access_count)'
            ),
            'get': 'access_count = access_count + 1, hit_count = hit_count + 1',
            'cull': 'SELECT {fields} FROM "' + table_name + '" ORDER BY access_count LIMIT ?',
        },
    }

def format_table_name(table_name: str) -> str:
    """
    Formats the table name

    >>> format_table_name('my.test.table_name') 
    'my_test_tableName'
    """
    parts = table_name.split('.')
    parts = [to_camel(part) for part in parts]
    return '_'.join(parts)


class BaseSqliteDB(abc.ABC):
    """
    The Sqlite Database
    """
    timeout: Optional[float] = 60.0
    _db_kind: Optional[str] = 'SQLite'

    def __init__(
        self, 
        conn_uri: str, 
        table: str, 
        optimization: Optional[Optimization] = None, 
        timeout: Optional[float] = None,
        is_remote: Optional[bool] = False,
        **kwargs
    ):
        """
        Initializes the Sqlite Database
        """
        self._extra: t.Dict[str, t.Any] = {}
        self.conn_uri = conn_uri
        self.table = format_table_name(table)
        self.optimization = optimization
        self._io: t.Optional['Connection'] = None
        self._aio: t.Optional['AsyncConnection'] = None
        # self.configured: bool = kwargs.pop('configured', False)
        self.is_child_db = kwargs.pop('is_child_db', False)
        self._debug_enabled = kwargs.pop('debug_enabled', False)
        self.configured: bool = False
        self.is_remote = is_remote
        self._config = kwargs
        self._txn_id = None
        self._is_configuring: bool = False
        if timeout is not None: self.timeout = timeout
        self._policies = get_eviction_policies(table_name = self.table)
        self._child_kwargs = {
            'conn_uri': self.conn_uri,
            'optimization': self.optimization,
            'is_remote': self.is_remote,
            'timeout': self.timeout,
            'kwargs': self._config,
        }
        self._spawned: t.Dict[str, t.List[t.Union['Connection', 'AsyncConnection']]] = {
            'io': [],
            'aio': [],
        }
        self.pool = ThreadPool
        self._register_exit_()

    
    
    """
    The Sqlite Database Connection
    """

    def _get_io_(self, timeout: t.Optional[float] = None) -> 'Connection':
        """
        Returns the IO Connection
        """
        new = sqlite3.connect(
            self.conn_uri, 
            uri = self.is_remote, 
            timeout = timeout or (self.timeout if self.configured else 0.0),
            check_same_thread = False,
            isolation_level = None,
        )
        self._spawned['io'].append(new)
        return new

    async def _get_aio_(self, timeout: Optional[float] = None) -> 'aiosqlite.Connection':
        """
        [Async] Returns the IO Connection
        """
        new = await aiosqlite.connect(
            self.conn_uri,
            uri = self.is_remote,
            timeout = timeout or (self.timeout if self.configured else 0.0),
            check_same_thread = False,
            isolation_level = None,
        )
        self._spawned['aio'].append(new)
        return new
    

    def _init_io_(self) -> None:
        """
        Initializes the IO Connection
        """
        if self._io is not None: return
        self._io = self._get_io_()
        if self._debug_enabled: logger.info(f'[Sync] Opened {self._db_kind} Database Connection: {self.conn_uri}', prefix = self.table, colored = True)
        self._validate_table_(self._io)
        if not self.is_configured:
            self._is_configuring = True
            self.run_configure()
            self._io = self._get_io_()
            self._is_configuring = False

    async def _init_aio_(self) -> None:
        """
        [Async] Initializes the IO Connection
        """
        if self._aio is not None: return
        self._aio = await self._get_aio_()
        if self._debug_enabled: logger.info(f'[Async] Opened {self._db_kind} Database Connection: {self.conn_uri}', prefix = self.table, colored = True)
        await self._avalidate_table_()
        if not self.is_configured:
            self._is_configuring = True
            await self.arun_configure()
            self._aio = await self._get_aio_()
            self._is_configuring = False
            

    @property
    def io(self) -> 'Connection':
        """
        Returns the IO Connection
        """
        if self._io is None: self._init_io_()
        return self._io


    @property
    def aio(self) -> 'AsyncConnection':
        """
        [Async] Returns the IO Connection
        """
        if self._aio is None: raise ValueError('Async Connection must be configured first')
        return self._aio

    """
    IO Wrappers 
    """

    @property
    def _sql_debug(self) -> t.Callable[..., 'Cursor']:
        """
        Returns the SQL
        """
        def _execute(statement, *args, **kwargs):
            logger.info(f'SQL: {statement}: {args} {kwargs}')
            return self._io.execute(statement, *args, **kwargs)
        return _execute

    @property
    def _asql_debug(self) -> t.Callable[..., t.Awaitable['AsyncCursor']]:
        """
        [Async] Returns the SQL
        """
        async def _aexecute(statement, *args, **kwargs):
            logger.info(f'SQL: {statement}: {args} {kwargs}')
            return await self._aio.execute(statement, *args, **kwargs)
        return _aexecute
    

    @property
    def _sqlmany_debug(self) -> t.Callable[..., 'Cursor']:
        """
        Returns the SQL
        """
        def _executemany(statement, *args, **kwargs):
            logger.info(f'SQL: {statement}: {args} {kwargs}')
            return self._io.executemany(statement, *args, **kwargs)
        return _executemany

    @property
    def _asqlmany_debug(self) -> t.Callable[..., t.Awaitable['AsyncCursor']]:
        """
        [Async] Returns the SQL
        """
        async def _aexecutemany(statement, *args, **kwargs):
            logger.info(f'[Async] SQL: {statement}: {args} {kwargs}')
            return await self._aio.executemany(statement, *args, **kwargs)
        return _aexecutemany

    @property
    def _sql(self) -> t.Callable[..., 'Cursor']:
        """
        Returns the SQL
        """
        if not self._io and not self._is_configuring: self._init_io_()
        return self._sql_debug if self._debug_enabled else self._io.execute

    @property
    def _asql(self) -> t.Callable[..., t.Awaitable['AsyncCursor']]:
        """
        [Async] Returns the SQL
        """
        if self._aio or self._is_configuring:
            return self._asql_debug if self._debug_enabled else self._aio.execute
        else: raise RuntimeError('Async Connection must be configured first')

    
    @property
    def _sqlmany(self) -> t.Callable[..., 'Cursor']:
        """
        Returns the SQL
        """
        return self._sqlmany_debug if self._debug_enabled else self._io.executemany

    @property
    def _asqlmany(self) -> t.Callable[..., t.Awaitable['AsyncCursor']]:
        """
        [Async] Returns the SQL
        """
        return self._asqlmany_debug if self._debug_enabled else self._aio.executemany
    
    @property
    def _sql_retry(self) -> t.Callable[..., 'Cursor']:
        """
        Returns the SQL Retry
        """
        sql = self._sql
        def _execute_with_retry(statement, *args, **kwargs):
            start = time.time()
            while True:
                try:
                    return sql(statement, *args, **kwargs)
                except sqlite3.OperationalError as exc:
                    logger.info(f'{self._db_kind} Error: {exc}: {statement}')
                    if str(exc) != 'database is locked': raise
                    diff = time.time() - start
                    if diff > 60: raise
                    time.sleep(0.001)
        return _execute_with_retry

    
    @property
    def _asql_retry(self) -> t.Callable[..., t.Awaitable['AsyncCursor']]:
        """
        [Async] Returns the SQL Retry
        """
        sql = self._asql
        async def _aexecute_with_retry(statement, *args, **kwargs):
            start = time.time()
            while True:
                try:
                    return await sql(statement, *args, **kwargs)
                except aiosqlite.OperationalError as exc:
                    if str(exc) != 'database is locked': raise
                    diff = time.time() - start
                    if diff > 60: raise
                    await asyncio.sleep(0.001)
        return _aexecute_with_retry

    @property
    def _sqlmany_retry(self) -> t.Callable[..., 'Cursor']:
        """
        Returns the SQL Retry
        """
        sql = self._sqlmany
        def _executemany_with_retry(statement, *args, **kwargs):
            start = time.time()
            while True:
                try:
                    return sql(statement, *args, **kwargs)
                except sqlite3.OperationalError as exc:
                    logger.info(f'SQLite Error: {exc}: {statement}')
                    if str(exc) != 'database is locked': raise
                    diff = time.time() - start
                    if diff > 60: raise
                    time.sleep(0.001)
        return _executemany_with_retry

    
    @property
    def _asqlmany_retry(self) -> t.Callable[..., t.Awaitable['AsyncCursor']]:
        """
        [Async] Returns the SQL Retry
        """
        sql = self._asqlmany
        async def _aexecutemany_with_retry(statement, *args, **kwargs):
            start = time.time()
            while True:
                try:
                    return await sql(statement, *args, **kwargs)
                except aiosqlite.OperationalError as exc:
                    if str(exc) != 'database is locked': raise
                    diff = time.time() - start
                    if diff > 60: raise
                    await asyncio.sleep(0.001)
        return _aexecutemany_with_retry

    @property
    def _sql_io_(self) -> t.Dict[str, bool]:
        """
        Shows the current IO Connection
        """
        return {
            'aio': self._aio is not None,
            'io': self._io is not None,
        }
    
    def close(self):
        """
        Closes the IO Connection
        """
        if self._io is not None:
            self._io.close()
            self._io = None
    
    async def aclose(self):
        """
        Closes the IO Connection
        """
        if self._aio is not None:
            await self._aio.close()
            self._aio = None
        self.close()
    
    
    """
    Exit Methods
    """
    
    def _register_exit_(self):
        """
        Registers the exit functions
        """
        from lzo.utils.aexit import register
        with contextlib.suppress(Exception):
            register(self._aon_exit_)
        atexit.register(self._on_exit_)
    
    def _on_exit_(self):
        """
        Closes the DBs on Exit
        """
        self.close()
        self._close_conns()
    
    async def _aon_exit_(self):
        """
        Closes the DBs on Exit
        """
        await self.aclose()
        await self._aclose_conns()

    def _close_conns(self):
        """
        Close all connections
        """
        if not self._spawned['io']: return
        while self._spawned['io']:
            conn = self._spawned['io'].pop()
            with contextlib.suppress(Exception):
                conn.close()
    
    async def _aclose_conns(self):
        """
        Close all connections
        """
        if not self._spawned['aio']: return
        while self._spawned['aio']:
            conn = self._spawned['aio'].pop()
            if not conn._running: continue
            await conn.close()

    
    """
    Configuration Steps
    """

    def _check_table_configured_(self, conn: t.Optional['Connection'] = None) -> bool:
        # sourcery skip: extract-method
        """
        Checks whether the DB has configured tables
        """
        conn = conn or self._get_io_()
        try:
            result = conn.execute(f'SELECT EXISTS(SELECT 1 FROM sqlite_master WHERE type = "table" AND name = "{self.table}")').fetchone()
            return bool(result[0])
        except sqlite3.OperationalError: return False
    
    async def _acheck_table_configured_(self, conn: t.Optional['AsyncConnection'] = None) -> bool:
        # sourcery skip: extract-method
        """
        Checks whether the DB has configured tables
        """
        conn = conn or await self._get_aio_()
        try:
            result = await (await conn.execute(f'SELECT EXISTS(SELECT 1 FROM sqlite_master WHERE type = "table" AND name = "{self.table}")')).fetchone()
            configured = bool(result[0])
            if configured: self._extra['has_configured_table'] = configured
            return configured
        except aiosqlite.OperationalError: return False

    @eproperty
    def has_configured_table(self) -> bool:
        """
        Returns whether the DB has configured the primary table
        """
        return self._check_table_configured_()
    
    def _validate_table_settings_(self, conn: t.Optional['Connection'] = None) -> bool: 
        # sourcery skip: extract-method
        """
        Validates the table settings
        """
        conn = conn or self._get_io_()
        if not self.has_configured_table: return False
        try:
            update_statement = f'UPDATE "{self.table}_settings" SET value = ? WHERE key = ?'
            # This means that the DB should have already been configured
            current_settings = dict(conn.execute(f'SELECT key, value FROM "{self.table}_settings"').fetchall())
            settings = get_optimized_settings(optimization = self.optimization, **self._config)
            diff_settings = dict_diff(current_settings, settings)['value_diffs']
            if diff_settings:
                for key, value in diff_settings.items():
                    if key.startswith('sqlite_'): 
                        if self._debug_enabled: logger.info(f'Skipping Diff: {key}: {value}')
                        continue
                    if self._debug_enabled: logger.info(f'Settings Diff: {key}: {value}')
                    current_settings[key] = value[1]
                    conn.execute(update_statement, (value[1], key))
            # We'll just set the values
            self._extra.update(current_settings)

            # Create tag index if requested.
            if self.tag_index:  # pylint: disable=no-member
                self.create_tag_index()
            # else:
            #     self.drop_tag_index()
            self.configured = True
            return True
        except sqlite3.OperationalError: 
            return False

    async def _avalidate_table_settings_(self, conn: t.Optional['AsyncConnection'] = None) -> bool: 
        # sourcery skip: extract-method
        """
        Validates the table settings
        """
        conn = conn or await self._get_aio_()
        if not await self._acheck_table_configured_(conn): return False
        try:
            update_statement = f'UPDATE "{self.table}_settings" SET value = ? WHERE key = ?'
            # This means that the DB should have already been configured
            current_settings: t.Dict[str, t.Union[str, int]] = dict(await (await conn.execute(f'SELECT key, value FROM "{self.table}_settings" ')).fetchall())
            settings = get_optimized_settings(optimization = self.optimization, **self._config)
            diff_settings = dict_diff(current_settings, settings)['value_diffs']
            if diff_settings:
                for key, value in diff_settings.items():
                    if key.startswith('sqlite_'): 
                        if self._debug_enabled: logger.info(f'Skipping Diff: {key}: {value}')
                        continue
                    if self._debug_enabled: logger.info(f'Settings Diff: {key}: {value}')
                    current_settings[key] = value[1]
                    await conn.execute(update_statement, (value[1], key))
            # We'll just set the values

            self._extra.update(current_settings)
            # Create tag index if requested.
            if self.tag_index:  # pylint: disable=no-member
                await self.acreate_tag_index()
        
            self.configured = True
            self._extra['has_configured_settings'] = True
            return True
        except aiosqlite.OperationalError: 
            return False
    
    @eproperty
    def has_configured_settings(self) -> bool:
        """
        Returns whether the DB has configured settings
        """
        return self._validate_table_settings_()

    @eproperty
    def _page_size(self) -> int:
        """
        Returns the page size
        """
        conn = self._io or self._get_io_()
        ((page_size,),) = conn.execute('PRAGMA page_size').fetchall()
        return page_size
    
    @eproperty
    def is_configured(self) -> bool:
        """
        Returns whether the DB has been configured
        """
        return self.has_configured_table and self.has_configured_settings

    def _validate_table_(self, conn: t.Optional['Connection'] = None) -> bool:
        # sourcery skip: extract-method
        """
        Validates the table settings
        """
        if self._extra.get('is_configured', False): return
        conn = conn or self._get_io_()
        self._extra['has_configured_table'] = self._check_table_configured_(conn)
        self._extra['has_configured_settings'] = self._validate_table_settings_(conn)
    
    async def _avalidate_table_(self, conn: t.Optional['AsyncConnection'] = None) -> bool:
        # sourcery skip: extract-method
        """
        Validates the table settings
        """
        if self._extra.get('is_configured', False): return
        provided_conn = conn is not None
        conn = conn or await self._get_aio_()
        await self._acheck_table_configured_(conn)
        await self._avalidate_table_settings_(conn)
        ((page_size,),) = (await (await conn.execute('PRAGMA page_size')).fetchall())
        self._extra['_page_size'] = page_size
        if not provided_conn: await conn.close()


    """
    Dynamic Properties
    """

    def _get_default_property(self, key):
        """
        Gets a default property
        """
        if self._config.get(key): return self._config[key]
        settings = get_optimized_settings(optimization = self.optimization)
        return settings[key]


    @eproperty
    def statistics(self) -> t.Union[bool, int]:
        """
        Returns the statistics
        """
        return self.reset('statistics', is_getter = True)

    @eproperty
    def tag_index(self) -> t.Union[bool, int]:
        """
        Returns the tag index
        """
        return self.reset('tag_index', is_getter = True)

    @eproperty
    def cull_limit(self) -> int:
        """
        Returns the cull limit
        """
        return self.reset('cull_limit', is_getter = True)
    
    @eproperty
    def size_limit(self) -> int:
        """
        Returns the size limit of items in the DB
        """
        return self.reset('size_limit', is_getter = True)
    
    @eproperty
    def eviction_policy(self) -> str:
        """
        Returns the Eviction Policy
        """
        if not self.configured:
            return self._get_default_property('eviction_policy')
        return self.reset('eviction_policy', is_getter = True)

    
    @eproperty
    def eviction_policies(self) -> Dict[str, str]:
        """
        Returns the eviction policies
        """
        return self._policies[self.eviction_policy]

    def run_configure(self):
        """
        Runs the configure
        """
        sql = self._sql_retry
        try: current_settings = dict(sql(f'SELECT key, value FROM "{self.table}_settings" ').fetchall())
        except sqlite3.OperationalError: current_settings = {}

        settings = get_optimized_settings(optimization = self.optimization, **current_settings)
        if self._config: settings.update(self._config)

        for key in METADATA:
            settings.pop(key, None)

        # Chance to set pragmas before any tables are created.
        for key, value in sorted(settings.items()):
            if key.startswith('sqlite_'):
                self.reset(key, value, update=False)
        
        # Setup Settings table.
        sql(
            f'CREATE TABLE IF NOT EXISTS "{self.table}_settings" ('
            ' key TEXT NOT NULL UNIQUE,'
            ' value)'
        )

        # Set cached attributes: updates settings and sets pragmas.
        for key, value in settings.items():
            query = f'INSERT OR REPLACE INTO "{self.table}_settings" VALUES (?, ?)'
            sql(query, (key, value))
            self.reset(key, value)

        for key, value in METADATA.items():
            query = f'INSERT OR IGNORE INTO "{self.table}_settings" VALUES (?, ?)'
            sql(query, (key, value))
            self.reset(key)

        # Setup Cache table.
        sql(
            f'CREATE TABLE IF NOT EXISTS "{self.table}" ('
            ' rowid INTEGER PRIMARY KEY,'
            ' key BLOB UNIQUE,'
            ' value BLOB,'
            ' store_time REAL,'
            ' expire_time REAL,'
            ' access_time REAL,'
            ' access_count INTEGER DEFAULT 0,'
            ' hit_count INTEGER DEFAULT 0,' # This doesn't reset whenever it is updated
            ' tag BLOB,'
            ' size INTEGER DEFAULT 0)'
        )
        sql(
            f'CREATE UNIQUE INDEX IF NOT EXISTS "{self.table}_key_rowid" ON'
            f' "{self.table}"(key, rowid)'
        )
        sql(
            f'CREATE UNIQUE INDEX IF NOT EXISTS "{self.table}_key_hits" ON'
            f' "{self.table}"(key, hit_count)'
        )
        sql(
            f'CREATE INDEX IF NOT EXISTS "{self.table}_expire_time" ON'
            f' "{self.table}" (expire_time) WHERE expire_time IS NOT NULL'
        )
        query = self.eviction_policies['init']
        if query is not None: sql(query)

        # Use triggers to keep Metadata updated.
        sql(
            f'CREATE TRIGGER IF NOT EXISTS "{self.table}_settings_count_insert"'
            f' AFTER INSERT ON "{self.table}" FOR EACH ROW BEGIN'
            f' UPDATE "{self.table}_settings" SET value = value + 1'
            ' WHERE key = "count"; END'
        )

        sql(
            f'CREATE TRIGGER IF NOT EXISTS "{self.table}_settings_count_delete"'
            f' AFTER DELETE ON "{self.table}" FOR EACH ROW BEGIN'
            f' UPDATE "{self.table}_settings" SET value = value - 1'
            ' WHERE key = "count"; END'
        )

        sql(
            f'CREATE TRIGGER IF NOT EXISTS "{self.table}_settings_size_insert"'
            f' AFTER INSERT ON "{self.table}" FOR EACH ROW BEGIN'
            f' UPDATE "{self.table}_settings" SET value = value + NEW.size'
            ' WHERE key = "size"; END'
        )

        sql(
            f'CREATE TRIGGER IF NOT EXISTS "{self.table}_settings_size_update"'
            f' AFTER UPDATE ON "{self.table}" FOR EACH ROW BEGIN'
            f' UPDATE "{self.table}_settings"'
            ' SET value = value + NEW.size - OLD.size'
            ' WHERE key = "size"; END'
        )

        sql(
            f'CREATE TRIGGER IF NOT EXISTS "{self.table}_settings_size_delete"'
            f' AFTER DELETE ON "{self.table}" FOR EACH ROW BEGIN'
            f' UPDATE "{self.table}_settings" SET value = value - OLD.size'
            ' WHERE key = "size"; END'
        )


        # Create tag index if requested.
        if self.tag_index:  # pylint: disable=no-member
            self.create_tag_index()
        # else:
        #     self.drop_tag_index()

        self.configured = True
        self.is_configured = True

        # Close and re-open database connection with given timeout.
        self.close()
        if self._debug_enabled: logger.info(f'Completed SQLite Database Connection: {self.conn_uri}')
        # self._get_io_()

    async def arun_configure(self):
        """
        Runs the configure
        """
        sql = self._asql_retry
        try: current_settings: t.Dict[str, t.Union[str, int]] = dict(await (await sql(f'SELECT key, value FROM "{self.table}_settings" ')).fetchall())
        except aiosqlite.OperationalError: current_settings = {}

        settings = get_optimized_settings(optimization = self.optimization, **current_settings)
        if self._config: settings.update(self._config)

        for key in METADATA:
            settings.pop(key, None)

        # Chance to set pragmas before any tables are created.
        for key, value in sorted(settings.items()):
            if key.startswith('sqlite_'):
                await self.areset(key, value, update=False)
        
        # Setup Settings table.
        await sql(
            f'CREATE TABLE IF NOT EXISTS "{self.table}_settings" ('
            ' key TEXT NOT NULL UNIQUE,'
            ' value)'
        )

        # Set cached attributes: updates settings and sets pragmas.
        for key, value in settings.items():
            query = f'INSERT OR REPLACE INTO "{self.table}_settings" VALUES (?, ?)'
            await sql(query, (key, value))
            await self.areset(key, value)

        for key, value in METADATA.items():
            query = f'INSERT OR IGNORE INTO "{self.table}_settings" VALUES (?, ?)'
            await sql(query, (key, value))
            await self.areset(key)

        # Setup Cache table.
        await sql(
            f'CREATE TABLE IF NOT EXISTS "{self.table}" ('
            ' rowid INTEGER PRIMARY KEY,'
            ' key BLOB UNIQUE,'
            ' value BLOB,'
            ' store_time REAL,'
            ' expire_time REAL,'
            ' access_time REAL,'
            ' access_count INTEGER DEFAULT 0,'
            ' hit_count INTEGER DEFAULT 0,' # This doesn't reset whenever it is updated
            ' tag BLOB,'
            ' size INTEGER DEFAULT 0)'
        )
        await sql(
            f'CREATE UNIQUE INDEX IF NOT EXISTS "{self.table}_key_rowid" ON'
            f' "{self.table}"(key, rowid)'
        )
        await sql(
            f'CREATE UNIQUE INDEX IF NOT EXISTS "{self.table}_key_hits" ON'
            f' "{self.table}"(key, hit_count)'
        )
        await sql(
            f'CREATE INDEX IF NOT EXISTS "{self.table}_expire_time" ON'
            f' "{self.table}" (expire_time) WHERE expire_time IS NOT NULL'
        )
        query = self.eviction_policies['init']
        if query is not None: await sql(query)

        # Use triggers to keep Metadata updated.
        await sql(
            f'CREATE TRIGGER IF NOT EXISTS "{self.table}_settings_count_insert"'
            f' AFTER INSERT ON "{self.table}" FOR EACH ROW BEGIN'
            f' UPDATE "{self.table}_settings" SET value = value + 1'
            ' WHERE key = "count"; END'
        )

        await sql(
            f'CREATE TRIGGER IF NOT EXISTS "{self.table}_settings_count_delete"'
            f' AFTER DELETE ON "{self.table}" FOR EACH ROW BEGIN'
            f' UPDATE "{self.table}_settings" SET value = value - 1'
            ' WHERE key = "count"; END'
        )

        await sql(
            f'CREATE TRIGGER IF NOT EXISTS "{self.table}_settings_size_insert"'
            f' AFTER INSERT ON "{self.table}" FOR EACH ROW BEGIN'
            f' UPDATE "{self.table}_settings" SET value = value + NEW.size'
            ' WHERE key = "size"; END'
        )

        await sql(
            f'CREATE TRIGGER IF NOT EXISTS {self.table}_settings_size_update'
            f' AFTER UPDATE ON "{self.table}" FOR EACH ROW BEGIN'
            f' UPDATE "{self.table}_settings"'
            ' SET value = value + NEW.size - OLD.size'
            ' WHERE key = "size"; END'
        )

        await sql(
            f'CREATE TRIGGER IF NOT EXISTS "{self.table}_settings_size_delete"'
            f' AFTER DELETE ON "{self.table}" FOR EACH ROW BEGIN'
            f' UPDATE "{self.table}_settings" SET value = value - OLD.size'
            ' WHERE key = "size"; END'
        )


        # Create tag index if requested.
        if self.tag_index:  # pylint: disable=no-member
            await self.acreate_tag_index()
        # else:
        #     self.drop_tag_index()

        self.configured = True
        self.is_configured = True

        # Close and re-open database connection with given timeout.
        await self.aclose()
        if self._debug_enabled: logger.info(f'Completed SQLite Database Connection: {self.conn_uri}')
        # self._get_io_()




    """
    Context Managers
    """

    @contextlib.contextmanager
    def transact(self, retry: bool = False) -> t.Generator[t.Callable[..., 'Cursor'], None, None]:
        """Context manager to perform a transaction by locking the cache.

        While the cache is locked, no other write operation is permitted.
        Transactions should therefore be as short as possible. Read and write
        operations performed in a transaction are atomic. Read operations may
        occur concurrent to a transaction.

        Transactions may be nested and may not be shared between threads.

        Raises :exc:`Timeout` error when database timeout occurs and `retry` is
        `False` (default).

        >>> cache = Cache()
        >>> with cache.transact():  # Atomically increment two keys.
        ...     _ = cache.incr('total', 123.4)
        ...     _ = cache.incr('count', 1)
        >>> with cache.transact():  # Atomically calculate average.
        ...     average = cache['total'] / cache['count']
        >>> average
        123.4

        :param bool retry: retry if database timeout occurs (default False)
        :return: context manager for use in `with` statement
        :raises Timeout: if database timeout occurs

        """
        with self._transact(retry=retry):
            yield

    @contextlib.contextmanager
    def _transact(self, retry: bool = False) -> t.Generator[t.Callable[..., 'Cursor'], None, None]:
        """
        Transaction Context Manager
        """
        self._init_io_()
        sql = self._sql
        # try:
        #     tid = asyncio.get_running_loop()._thread_id
        # except:
        tid = threading.get_ident()
        txn_id = self._txn_id
        if tid == txn_id: begin = False
        else:
            while True:
                try:
                    sql('BEGIN IMMEDIATE')
                    begin = True
                    self._txn_id = tid
                    break
                except sqlite3.OperationalError as e:
                    logger.info(f'SQLite Error: {e}')
                    if retry: continue
                    raise Timeout from None

        try:
            yield sql
        except BaseException:
            if begin:
                assert self._txn_id == tid
                self._txn_id = None
                sql('ROLLBACK')
            raise
        else:
            if begin:
                assert self._txn_id == tid
                self._txn_id = None
                sql('COMMIT')


    @contextlib.asynccontextmanager
    async def atransact(self, retry: bool = False) -> t.AsyncGenerator[t.Awaitable[t.Callable[..., t.Awaitable['AsyncCursor']]], None]:
        """Context manager to perform a transaction by locking the cache.

        While the cache is locked, no other write operation is permitted.
        Transactions should therefore be as short as possible. Read and write
        operations performed in a transaction are atomic. Read operations may
        occur concurrent to a transaction.

        Transactions may be nested and may not be shared between threads.

        Raises :exc:`Timeout` error when database timeout occurs and `retry` is
        `False` (default).

        >>> cache = Cache()
        >>> with cache.transact():  # Atomically increment two keys.
        ...     _ = cache.incr('total', 123.4)
        ...     _ = cache.incr('count', 1)
        >>> with cache.transact():  # Atomically calculate average.
        ...     average = cache['total'] / cache['count']
        >>> average
        123.4

        :param bool retry: retry if database timeout occurs (default False)
        :return: context manager for use in `with` statement
        :raises Timeout: if database timeout occurs

        """
        async with self._atransact(retry=retry):
            yield

    @contextlib.asynccontextmanager
    async def _atransact(self, retry: bool = False) -> t.AsyncGenerator[t.Callable[..., t.Awaitable['AsyncCursor']], None]:
        """
        [Async] Transaction Context Manager
        """
        await self._init_aio_()
        sql = self._asql
        tid = asyncio.get_running_loop()._thread_id
        txn_id = self._txn_id
        if tid == txn_id: begin = False
        else:
            while True:
                try:
                    await sql('BEGIN IMMEDIATE')
                    begin = True
                    self._txn_id = tid
                    break
                except aiosqlite.OperationalError:
                    if retry: continue
                    raise Timeout from None

        try:
            yield sql
        except BaseException:
            if begin:
                assert self._txn_id == tid
                self._txn_id = None
                await sql('ROLLBACK')
            raise
        else:
            if begin:
                assert self._txn_id == tid
                self._txn_id = None
                await sql('COMMIT')

    @contextlib.contextmanager
    def _transact_many(self, retry: bool = False) -> t.Generator[t.Callable[..., t.Tuple['Cursor', 'Cursor']], None, None]:
        """
        [Many] Transaction Context Manager

        Returns a tuple of (io.execute, io.executemany) cursors
        """
        self._init_io_()
        sql = self._sql
        sqlmany = self._sqlmany
        tid = threading.get_ident()
        txn_id = self._txn_id
        if tid == txn_id: begin = False
        else:
            while True:
                try:
                    sql('BEGIN IMMEDIATE')
                    begin = True
                    self._txn_id = tid
                    break
                except sqlite3.OperationalError as e:
                    logger.info(f'SQLite Error: {e}')
                    if retry: continue
                    raise Timeout from None

        try:
            yield (sql, sqlmany)
        except BaseException:
            if begin:
                assert self._txn_id == tid
                self._txn_id = None
                sql('ROLLBACK')
            raise
        else:
            if begin:
                assert self._txn_id == tid
                self._txn_id = None
                sql('COMMIT')


    @contextlib.asynccontextmanager
    async def _atransact_many(self, retry: bool = False) -> t.AsyncGenerator[t.Callable[..., t.Tuple[t.Awaitable['AsyncCursor'], t.Awaitable['AsyncCursor']]], None]:
        """
        [Async - Many] Transaction Context Manager
        
        Returns a tuple of (aio.execute, aio.executemany) cursors
        """
        await self._init_aio_()
        sql = self._asql
        sqlmany = self._asqlmany
        tid = asyncio.get_running_loop()._thread_id
        txn_id = self._txn_id
        if tid == txn_id: begin = False
        else:
            while True:
                try:
                    await sql('BEGIN IMMEDIATE')
                    begin = True
                    self._txn_id = tid
                    break
                except aiosqlite.OperationalError:
                    if retry: continue
                    raise Timeout from None

        try:
            yield (sql, sqlmany)
        except BaseException:
            if begin:
                assert self._txn_id == tid
                self._txn_id = None
                await sql('ROLLBACK')
            raise
        else:
            if begin:
                assert self._txn_id == tid
                self._txn_id = None
                await sql('COMMIT')
    
    def _format_keys_(self, keys: t.Iterable[str]) -> str:
        """
        Formats the keys
    
        >>> format_keys_(['key1', 'key2'])
        '"key1", "key2"'
        """
        return ', '.join([f'"{key}"' for key in keys])
    
    def _format_rowids_(self, rowids: t.Iterable[int]) -> str:
        """
        Formats the rowids
        
        >>> format_rowids_([1, 2])
        '(1, 2)'
        """
        return  ', '.join(str(rowid) for rowid in rowids)



    """
    Basic CRUD Methods 
    """
    
    def _select_delete(
        self, 
        select: str, 
        args: List[Any],
        row_index: int = 0,
        arg_index: int = 0,
        retry: bool = False
    ):
        """
        Delete rows from Cache
        """
        count = 0
        delete = f'DELETE FROM "{self.table}" WHERE rowid IN (%s)'
        try:
            while True:
                with self._transact(retry) as sql:
                    rows = sql(select, args).fetchall()
                    if not rows: break
                    count += len(rows)
                    sql(delete % self._format_rowids_((row[0] for row in rows)))
                    # sql(delete % ','.join(str(row[0]) for row in rows))
                    for row in rows:
                        args[arg_index] = row[row_index]
                    #     cleanup(row[-1])

        except Timeout: raise Timeout(count) from None
        return count
    
    async def _aselect_delete(
        self, 
        select: str, 
        args: List[Any],
        row_index: int = 0,
        arg_index: int = 0,
        retry: bool = False
    ):
        """
        Delete rows from Cache
        """
        count = 0
        delete = f'DELETE FROM "{self.table}" WHERE rowid IN (%s)'
        try:
            while True:
                async with self._atransact(retry) as sql:
                    rows = await (await sql(select, args)).fetchall()
                    if not rows: break
                    count += len(rows)
                    await sql(delete % self._format_rowids_((row[0] for row in rows)))
                    # await sql(delete % ','.join(str(row[0]) for row in rows))
                    for row in rows:
                        args[arg_index] = row[row_index]
        
        except Timeout: raise Timeout(count) from None
        return count


    def _row_update(self, rowid: int, now: float, columns: t.Tuple[t.Any, ...], is_async: bool = False):
        """
        Updates the row
        """
        sql = self._asql if is_async else self._sql
        # expire_time, tag, size, mode, filename, value = columns
        expire_time, tag, size, value = columns
        return sql(
            f'UPDATE "{self.table}" SET'
            ' store_time = ?,'
            ' expire_time = ?,'
            ' access_time = ?,'
            ' access_count = ?,'
            ' tag = ?,'
            ' size = ?,'
            ' value = ?'
            ' WHERE rowid = ?',
            (
                now,  # store_time
                expire_time,
                now,  # access_time
                0,  # access_count
                tag,
                size,
                value,
                rowid,
            ),
        )


    # def _row_insert(self, key: str, raw: int, now: float, columns: t.Tuple[t.Any, ...], is_async: bool = False):
    def _row_insert(self, key: str, now: float, columns: t.Tuple[t.Any, ...], is_async: bool = False):
        """
        Inserts the row
        """
        sql = self._asql if is_async else self._sql
        # expire_time, tag, size, mode, filename, value = columns
        expire_time, tag, size, value = columns
        return sql(
            f'INSERT INTO "{self.table}"('
            ' key, store_time, expire_time, access_time,'
            ' access_count, tag, size, value'
            ') VALUES (?, ?, ?, ?, ?, ?, ?, ?)',
            (
                key,
                now,  # store_time
                expire_time,
                now,  # access_time
                0,  # access_count
                tag,
                size,
                value,
            ),
        )

    def _row_upsert(self, key: str, now: float, columns: t.Tuple[t.Any, ...], is_async: bool = False, sql: t.Union[t.Callable[..., t.Awaitable['Cursor']], t.Callable[..., 'Cursor']] = None):
        """
        Upserts a row
        """
        if sql is None: sql = self._asql if is_async else self._sql
        expire_time, tag, size, value = columns
        return sql(
            f'INSERT INTO "{self.table}" ('
            ' key, store_time, expire_time, access_time,'
            ' access_count, tag, size, value'
            ') VALUES (?, ?, ?, ?, ?, ?, ?, ?)'
            ' ON CONFLICT(key) DO UPDATE SET'
                ' store_time = excluded.store_time,'
                ' expire_time = excluded.expire_time,'
                ' access_time = excluded.access_time,'
                ' access_count = excluded.access_count,'
                ' tag = excluded.tag,'
                ' size = excluded.size,'
                ' value = excluded.value',
            (
                key,
                now,  # store_time
                expire_time,
                now,  # access_time
                0,  # access_count
                tag,
                size,
                value,
            ),
        )

    def _bulk_insert(self, data: t.Dict[str, t.Any], now: float, columns: t.Tuple[t.Any, ...], is_async: bool = False, sql: t.Union[t.Callable[..., t.Awaitable['Cursor']], t.Callable[..., 'Cursor']] = None):
        """
        Inserts the many rows of data
        where data = {
            key: value
        }
        """
        if sql is None: sql = self._asql if is_async else self._sql
        # sql = self._asql if is_async else self._sql
        # expire_time, tag, size, mode, filename, value = columns
        expire_time, tag = columns
        batch_data = [
            (
                key,
                now,  # store_time
                expire_time,
                now,  # access_time
                0,  # access_count
                tag,
                self.get_object_size(value),
                value,
            )
            for key, value in data.items()
        ]
        if not batch_data: return
        sql(
            f'INSERT INTO "{self.table}"('
            ' key, store_time, expire_time, access_time,'
            ' access_count, tag, size, value'
            ') VALUES (?, ?, ?, ?, ?, ?, ?, ?)',
            batch_data,
        )

    def _bulk_update(self, data: t.Dict[str, t.Any], now: float, columns: t.Tuple[t.Any, ...], is_async: bool = False):
        """
        Updates the a batch of rows of data
        """
        sql = self._asqlmany if is_async else self._sqlmany
        expire_time, tag = columns
        batch_data = [
            (
                now,  # store_time
                expire_time,
                now,  # access_time
                0,  # access_count
                tag,
                self.get_object_size(value),
                value,
                key,
            )
            for key, value in data.items()
        ]
        return sql(
            f'UPDATE "{self.table}" SET'
            ' store_time = ?,'
            ' expire_time = ?,'
            ' access_time = ?,'
            ' access_count = ?,'
            ' tag = ?,'
            ' size = ?,'
            ' value = ?'
            ' WHERE key = ?',
            batch_data,
        )

    def _bulk_upsert(self, data: t.Dict[str, t.Any], now: float, columns: t.Tuple[t.Any, ...], is_async: bool = False, sql: t.Callable[..., t.Awaitable[t.Any]] = None):
        """
        Upserts the a batch of rows of data
        """
        if sql is None: sql = self._asqlmany if is_async else self._sqlmany
        expire_time, tag = columns
        batch_data = [
            (
                key,
                now,  # store_time
                expire_time,
                now,  # access_time
                0,  # access_count
                tag,
                self.get_object_size(value),
                value,
            )
            for key, value in data.items()
        ]
        return sql(
            f'INSERT INTO "{self.table}" ('
            ' key, store_time, expire_time, access_time,'
            ' access_count, tag, size, value'
            ') VALUES (?, ?, ?, ?, ?, ?, ?, ?)'
            ' ON CONFLICT(key) DO UPDATE SET'
                ' store_time = excluded.store_time,'
                ' expire_time = excluded.expire_time,'
                ' access_time = excluded.access_time,'
                ' access_count = excluded.access_count,'
                ' tag = excluded.tag,'
                ' size = excluded.size,'
                ' value = excluded.value',
            batch_data,
        )
    
    """
    Index / Tag Methods
    """

    def create_tag_index(self):
        """Create tag index on cache database.

        It is better to initialize cache with `tag_index=True` than use this.

        :raises Timeout: if database timeout occurs

        """
        sql = self._sql
        sql(
            f'CREATE INDEX IF NOT EXISTS "{self.table}_tag_rowid" ON "{self.table}"(tag, rowid) '
            'WHERE tag IS NOT NULL'
        )
        self.reset('tag_index', 1)
    
    async def acreate_tag_index(self):
        """Create tag index on cache database.

        It is better to initialize cache with `tag_index=True` than use this.

        :raises Timeout: if database timeout occurs

        """
        sql = self._asql
        await sql(
            f'CREATE INDEX IF NOT EXISTS "{self.table}_tag_rowid" ON "{self.table}"(tag, rowid) '
            'WHERE tag IS NOT NULL'
        )
        await self.areset('tag_index', 1)

    def drop_tag_index(self):
        """Drop tag index on cache database.

        :raises Timeout: if database timeout occurs
        """
        sql = self._sql
        sql(f'DROP INDEX IF EXISTS "{self.table}_tag_rowid"')
        self.reset('tag_index', 0)
    
    async def adrop_tag_index(self):
        """Drop tag index on cache database.

        :raises Timeout: if database timeout occurs
        """
        await self._asql(f'DROP INDEX IF EXISTS "{self.table}_tag_rowid"')
        await self.areset('tag_index', 0)
    

    def get_object_size(self, value: t.Any) -> int:
        """
        Returns the object size
        """
        return len(value) if isinstance(value, (bytes, bytearray)) else len(pickle.dumps(value))

    def reset(self, key: str, value: t.Any = ENOVAL, update: bool = True, is_getter: bool = False):
        # sourcery skip: extract-duplicate-method
        """Reset `key` and `value` item from Settings table.

        Use `reset` to update the value of Cache settings correctly. Cache
        settings are stored in the Settings table of the SQLite database. If
        `update` is ``False`` then no attempt is made to update the database.

        If `value` is not given, it is reloaded from the Settings
        table. Otherwise, the Settings table is updated.

        Settings with the ``disk_`` prefix correspond to Disk
        attributes. Updating the value will change the unprefixed attribute on
        the associated Disk instance.

        Settings with the ``sqlite_`` prefix correspond to SQLite
        pragmas. Updating the value will execute the corresponding PRAGMA
        statement.

        SQLite PRAGMA statements may be executed before the Settings table
        exists in the database by setting `update` to ``False``.

        :param str key: Settings key for item
        :param value: value for item (optional)
        :param bool update: update database Settings table (default True)
        :return: updated value for item
        :raises Timeout: if database timeout occurs

        """
        sql = self._sql
        sql_retry = self._sql_retry
        if value is ENOVAL:
            select = f'SELECT value FROM "{self.table}_settings"  WHERE key = ?'
            ((value,),) = sql_retry(select, (key,)).fetchall()
            if not is_getter: self._extra[key] = value
            if key == 'eviction_policy': self._extra['eviction_policies'] = self._policies[value]
            return value

        if update and value:
            statement = f'UPDATE "{self.table}_settings" SET value = ? WHERE key = ?'
            sql_retry(statement, (value, key))

        if key.startswith('sqlite_') and not self.is_child_db:
            # if self.is_child_db: return
            pragma = key[7:]

            # 2016-02-17 GrantJ - PRAGMA and isolation_level=None
            # don't always play nicely together. Retry setting the
            # PRAGMA. I think some PRAGMA statements expect to
            # immediately take an EXCLUSIVE lock on the database. I
            # can't find any documentation for this but without the
            # retry, stress will intermittently fail with multiple
            # processes.

            # 2018-11-05 GrantJ - Avoid setting pragma values that
            # are already set. Pragma settings like auto_vacuum and
            # journal_mode can take a long time or may not work after
            # tables have been created.

            start = time.time()
            while True:
                try:
                    try:
                        ((old_value,),) = sql(f'PRAGMA {pragma}').fetchall()
                        update = old_value != value
                    except ValueError: update = True
                    if update: sql(f'PRAGMA {pragma} = {value}').fetchall()
                    break
                except sqlite3.OperationalError as exc:
                    if str(exc) != 'database is locked': raise
                    diff = time.time() - start
                    if diff > 60: raise
                    time.sleep(0.001)

        if not is_getter: self._extra[key] = value
        if key == 'eviction_policy': self._extra['eviction_policies'] = self._policies[value]
        return value
    

    async def areset(self, key: str, value: t.Any = ENOVAL, update: bool = True, is_getter: bool = False):
        # sourcery skip: extract-duplicate-method
        """Reset `key` and `value` item from Settings table.

        Use `reset` to update the value of Cache settings correctly. Cache
        settings are stored in the Settings table of the SQLite database. If
        `update` is ``False`` then no attempt is made to update the database.

        If `value` is not given, it is reloaded from the Settings
        table. Otherwise, the Settings table is updated.

        Settings with the ``disk_`` prefix correspond to Disk
        attributes. Updating the value will change the unprefixed attribute on
        the associated Disk instance.

        Settings with the ``sqlite_`` prefix correspond to SQLite
        pragmas. Updating the value will execute the corresponding PRAGMA
        statement.

        SQLite PRAGMA statements may be executed before the Settings table
        exists in the database by setting `update` to ``False``.

        :param str key: Settings key for item
        :param value: value for item (optional)
        :param bool update: update database Settings table (default True)
        :return: updated value for item
        :raises Timeout: if database timeout occurs

        """
        sql = self._asql
        sql_retry = self._asql_retry
        if value is ENOVAL:
            select = f'SELECT value FROM "{self.table}_settings"  WHERE key = ?'
            ((value,),) = await (await sql_retry(select, (key,))).fetchall()
            if not is_getter: self._extra[key] = value
            if key == 'eviction_policy': self._extra['eviction_policies'] = self._policies[value]
            return value

        if update and value:
            statement = f'UPDATE "{self.table}_settings" SET value = ? WHERE key = ?'
            await sql_retry(statement, (value, key))

        if key.startswith('sqlite_') and not self.is_child_db:
            # if self.is_child_db: return
            pragma = key[7:]

            start = time.time()
            while True:
                try:
                    try:
                        ((old_value,),) = await (await sql(f'PRAGMA {pragma}')).fetchall()
                        update = old_value != value
                    except ValueError: update = True
                    if update: await (await sql(f'PRAGMA {pragma} = {value}')).fetchall()
                    break
                except aiosqlite.OperationalError as exc:
                    if str(exc) != 'database is locked': raise
                    diff = time.time() - start
                    if diff > 60: raise
                    await asyncio.sleep(0.001)

        if not is_getter: self._extra[key] = value
        if key == 'eviction_policy': self._extra['eviction_policies'] = self._policies[value]
        return value
    
    def stats(self, enable=True, reset=False):
        """Return cache statistics hits and misses.

        :param bool enable: enable collecting statistics (default True)
        :param bool reset: reset hits and misses to 0 (default False)
        :return: (hits, misses)

        """
        # pylint: disable=E0203,W0201
        result = (self.reset('hits'), self.reset('misses'))
        if reset:
            self.reset('hits', 0)
            self.reset('misses', 0)

        self.reset('statistics', enable)
        return result

    async def astats(self, enable=True, reset=False):
        """Return cache statistics hits and misses.

        :param bool enable: enable collecting statistics (default True)
        :param bool reset: reset hits and misses to 0 (default False)
        :return: (hits, misses)

        """
        # pylint: disable=E0203,W0201
        hits, misses = await self.areset('hits'), await self.areset('misses')
        result = (hits, misses)
        if reset:
            await self.areset('hits', 0)
            await self.areset('misses', 0)
        
        await self.areset('statistics', enable)
        return result

    def volume(self):
        """Return estimated total size of cache on disk.

        :return: size in bytes
        """
        ((page_count,),) = self._sql('PRAGMA page_count').fetchall()
        return self._page_size * page_count + self.reset('size')
    

    async def avolume(self):
        """Return estimated total size of cache on disk.

        :return: size in bytes
        """
        ((page_count,),) = await (await self._asql('PRAGMA page_count')).fetchall()
        return self._page_size * page_count + await self.areset('size')


    def fetch_keys(
        self,
        pattern: t.Optional[str] = None,
        include_rowid: Optional[bool] = False,
        order: t.Optional[t.Literal['ASC', 'DESC']] = 'ASC',
        order_by: t.Optional[str] = 'rowid',
        limit: Optional[int] = None,
        retry: bool = False,
    ) -> t.Union[List[str], t.Tuple[List[str], List[int]]]:
        # sourcery skip: extract-method
        """
        Fetches the keys
        """
        select = (
            f'SELECT key, rowid FROM "{self.table}"'
        )
        args = ()
        if pattern: 
            select += ' WHERE key LIKE ?'
            args = (pattern,)
        if order: select += f' ORDER BY {order_by} {order}'
        if limit: select += f' LIMIT {limit}'
        with self._transact(retry) as sql:
            rows = sql(
                select,
                args,
            ).fetchall()
            if not rows: return [], [] if include_rowid else []
            keys = [row[0] for row in rows]
            if not include_rowid: return keys
            rowids = [row[1] for row in rows]
            return keys, rowids

    async def afetch_keys(
        self,
        pattern: t.Optional[str] = None,
        include_rowid: Optional[bool] = False,
        order: t.Optional[t.Literal['ASC', 'DESC']] = 'ASC',
        order_by: t.Optional[str] = 'rowid',
        limit: Optional[int] = None,
        retry: bool = False,
    ) -> t.Union[List[str], t.Tuple[List[str], List[int]]]:
        # sourcery skip: extract-method
        """
        Fetches the keys
        """
        select = (
            f'SELECT key, rowid FROM "{self.table}"'
        )
        args = ()
        if pattern: 
            select += ' WHERE key LIKE ?'
            args = (pattern,)
        if order: select += f' ORDER BY {order_by} {order}'
        if limit: select += f' LIMIT {limit}'
        
        async with self._atransact(retry) as sql:
            rows = await (await sql(
                select,
                args,
            )).fetchall()
            if not rows: return [], [] if include_rowid else []
            keys = [row[0] for row in rows]
            if not include_rowid: return keys
            rowids = [row[1] for row in rows]
            return keys, rowids
    """
    Class Methods
    """

    def _delete_key_(self, key: str, retry: bool = True):
        """Delete corresponding item for `key` from cache.

        Raises :exc:`Timeout` error when database timeout occurs and `retry` is
        `False` (default `True`).

        :param key: key matching item
        :param bool retry: retry if database timeout occurs (default True)
        :raises KeyError: if key is not found
        :raises Timeout: if database timeout occurs

        """
        with self._transact(retry) as sql:
            rows = sql(
                f'SELECT rowid FROM "{self.table}"'
                ' WHERE key = ? AND raw = ?'
                ' AND (expire_time IS NULL OR expire_time > ?)',
                (key, time.time()),
            ).fetchall()
            if not rows: raise KeyError(key)
            ((rowid,),) = rows
            sql(f'DELETE FROM "{self.table}" WHERE rowid = ?', (rowid,))
            return True
    
    async def _adelete_key_(self, key: str, retry: bool = True):
        """[Async Background] Delete corresponding item for `key` from cache.

        Raises :exc:`Timeout` error when database timeout occurs and `retry` is
        `False` (default `True`).

        :param key: key matching item
        :param bool retry: retry if database timeout occurs (default True)
        :raises KeyError: if key is not found
        :raises Timeout: if database timeout occurs

        """
        async with self._atransact(retry) as sql:
            rows = await (await sql(
                f'SELECT rowid FROM "{self.table}"'
                ' WHERE key = ?'
                ' AND (expire_time IS NULL OR expire_time > ?)',
                (key, time.time()),
            )).fetchall()
            if not rows: raise KeyError(key)
            ((rowid,),) = rows
            await sql(f'DELETE FROM "{self.table}" WHERE rowid = ?', (rowid,))
            return True

    def _delete_keys_(self, *keys: str, retry: bool = True) -> int:
        """Delete corresponding item for `key` from cache.

        Raises :exc:`Timeout` error when database timeout occurs and `retry` is
        `False` (default `True`).

        :param keys: keys matching item
        :param bool retry: retry if database timeout occurs (default True)
        :raises KeyError: if key is not found
        :raises Timeout: if database timeout occurs

        """
        with self._transact(retry) as sql:
            rows = sql(
                f'SELECT rowid FROM "{self.table}"'
                ' WHERE key IN (%s)'
                ' AND (expire_time IS NULL OR expire_time > ?)' % self._format_keys_(keys),
                (time.time(),),
            ).fetchall()
            if not rows: raise KeyError(keys)
            rowids = [row[0] for row in rows if row]
            sql(f'DELETE FROM "{self.table}" WHERE rowid IN (%s)' % self._format_rowids_(rowids))
            return len(rowids)
    
    async def _adelete_keys_(self, *keys: str, retry: bool = True) -> int:
        """[Async] Delete corresponding item for `keys` from cache.

        Raises :exc:`Timeout` error when database timeout occurs and `retry` is
        `False` (default `True`).

        :param keys: keys matching item
        :param bool retry: retry if database timeout occurs (default True)
        :raises KeyError: if key is not found
        :raises Timeout: if database timeout occurs
        """
        async with self._atransact(retry) as sql:
            rows = await (await sql(
                f'SELECT rowid FROM "{self.table}"'
                ' WHERE key IN (%s)'
                ' AND (expire_time IS NULL OR expire_time > ?)' % self._format_keys_(keys),
                (time.time(),),
            )).fetchall()
            if not rows: raise KeyError(keys)
            rowids = [row[0] for row in rows if row]
            await sql(f'DELETE FROM "{self.table}" WHERE rowid IN (%s)' % self._format_rowids_(rowids))
            return len(rowids)

    def _set_key_expiration_(self, key: str, ex: float, retry: bool = True) -> None:
        """
        Sets the expiration for a key
        """
        sql = self._sql_retry if retry else self._sql
        now = time.time()
        if ex < now: ex = now + ex
        sql(f'UPDATE "{self.table}" SET expire_time = ? WHERE key = ?', (ex, key))

    async def _aset_key_expiration_(self, key: str, ex: float, retry: bool = True) -> None:
        """
        [Async] Sets the expiration for a key
        """
        sql = self._asql_retry if retry else self._asql
        now = time.time()
        if ex < now: ex = now + ex
        await sql(f'UPDATE "{self.table}" SET expire_time = ? WHERE key = ?', (ex, key))

    def _set_key_tag_(self, key: str, tag: str, retry: bool = True) -> None:
        """
        Sets the tag for a key
        """
        sql = self._sql_retry if retry else self._sql
        sql(f'UPDATE "{self.table}" SET tag = ? WHERE key = ?', (tag, key))

    async def _aset_key_tag_(self, key: str, tag: str, retry: bool = True) -> None:
        """
        [Async] Sets the tag for a key
        """
        sql = self._asql_retry if retry else self._asql
        await sql(f'UPDATE "{self.table}" SET tag = ? WHERE key = ?', (tag, key))
    
    def contains(self, key: str, retry: bool = True) -> bool:
        """Return `True` if `key` matching item is found in cache.

        :param key: key matching item
        :return: True if key matching item

        """
        sql = self._sql_retry if retry else self._sql
        select = (
            f'SELECT rowid FROM "{self.table}"'
            ' WHERE key = ?'
            ' AND (expire_time IS NULL OR expire_time > ?)'
        )
        rows = sql(select, (key, time.time())).fetchall()
        return bool(rows)
    
    async def acontains(self, key: str, retry: bool = True) -> bool:
        """[Async Background] Return `True` if `key` matching item is found in cache.

        :param key: key matching item
        :return: True if key matching item

        """
        sql = self._asql_retry if retry else self._asql
        select = (
            f'SELECT rowid FROM "{self.table}"'
            ' WHERE key = ?'
            ' AND (expire_time IS NULL OR expire_time > ?)'
        )
        rows = await (await sql(select, (key, time.time()))).fetchall()
        return bool(rows)

    def _iter(self, ascending: bool = True) -> t.Generator[t.Any, None, None]:
        """
        Iterate Cache items in database sort order.
        """
        sql = self._sql
        rows = sql(f'SELECT MAX(rowid) FROM "{self.table}"').fetchall()
        ((max_rowid,),) = rows
        yield  # Signal ready.

        if max_rowid is None: return
        bound = max_rowid + 1
        limit = 100
        # _disk_get = self._disk.get
        rowid = 0 if ascending else bound
        select = (
            f'SELECT rowid, key FROM "{self.table}"'
            ' WHERE ? < rowid AND rowid < ?'
            ' ORDER BY rowid %s LIMIT ?'
        ) % ('ASC' if ascending else 'DESC')

        while True:
            args = (rowid, bound, limit) if ascending else (0, rowid, limit)
            rows = sql(select, args).fetchall()
            if not rows: break
            for rowid, key in rows: yield iter(key)
    
    async def _aiter(self, ascending: bool = True) -> t.AsyncGenerator[t.Any, None]:
        """
        [Async Background] Iterate Cache items in database sort order.
        """
        sql = self._asql
        rows = await (await sql(f'SELECT MAX(rowid) FROM "{self.table}"')).fetchall()
        ((max_rowid,),) = rows
        yield  # Signal ready.

        if max_rowid is None: return
        bound = max_rowid + 1
        limit = 100
        rowid = 0 if ascending else bound
        select = (
            f'SELECT rowid, key FROM "{self.table}"'
            ' WHERE ? < rowid AND rowid < ?'
            ' ORDER BY rowid %s LIMIT ?'
        ) % ('ASC' if ascending else 'DESC')

        while True:
            args = (rowid, bound, limit) if ascending else (0, rowid, limit)
            rows = await (await sql(select, args)).fetchall()
            if not rows: break
            for rowid, key in rows: yield aiter(key)

    
    def iterkeys(self, reverse=False):
        """Iterate Cache keys in database sort order.

        >>> cache = Cache()
        >>> for key in [4, 1, 3, 0, 2]:
        ...     cache[key] = key
        >>> list(cache.iterkeys())
        [0, 1, 2, 3, 4]
        >>> list(cache.iterkeys(reverse=True))
        [4, 3, 2, 1, 0]

        :param bool reverse: reverse sort order (default False)
        :return: iterator of Cache keys

        """
        sql = self._sql
        limit = 100
        if reverse:
            select = (
                f'SELECT key FROM "{self.table}"'
                ' ORDER BY key DESC LIMIT 1'
            )
            iterate = (
                f'SELECT key FROM "{self.table}"'
                ' WHERE key = ? OR key < ?'
                ' ORDER BY key DESC LIMIT ?'
            )
        else:
            select = (
                f'SELECT key FROM "{self.table}"'
                ' ORDER BY key ASC LIMIT 1'
            )
            iterate = (
                f'SELECT key FROM "{self.table}"'
                ' WHERE key = ? OR key > ?'
                ' ORDER BY key ASC LIMIT ?'
            )

        row = sql(select).fetchall()
        if row: ((key,),) = row
        else: return
        yield iter(key)
        while True:
            rows = sql(iterate, (key, key, limit)).fetchall()
            if not rows: break
            for key in rows: yield iter(key)
    
    async def aiterkeys(self, reverse=False):
        """[Async] Iterate Cache keys in database sort order.

        >>> cache = Cache()
        >>> for key in [4, 1, 3, 0, 2]:
        ...     cache[key] = key
        >>> list(cache.aiterkeys())
        [0, 1, 2, 3, 4]
        >>> list(cache.aiterkeys(reverse=True))
        [4, 3, 2, 1, 0]

        :param bool reverse: reverse sort order (default False)
        :return: iterator of Cache keys

        """
        sql = self._asql
        limit = 100
        if reverse:
            select = (
                f'SELECT key FROM "{self.table}"'
                ' ORDER BY key DESC LIMIT 1'
            )
            iterate = (
                f'SELECT key FROM "{self.table}"'
                ' WHERE key = ? OR key < ?'
                ' ORDER BY key DESC LIMIT ?'
            )
        else:
            select = (
                f'SELECT key FROM "{self.table}"'
                ' ORDER BY key ASC LIMIT 1'
            )
            iterate = (
                f'SELECT key FROM "{self.table}"'
                ' WHERE key = ? OR key > ?'
                ' ORDER BY key ASC LIMIT ?'
            )

        row = await (await sql(select)).fetchall()
        if row: ((key,),) = row
        else: return
        yield aiter(key)
        while True:
            rows = await (await sql(iterate, (key, key, limit))).fetchall()
            if not rows: break
            for key in rows: yield aiter(key)
    

    def __iter__(self):
        """Iterate keys in cache including expired items."""
        iterator = self._iter()
        next(iterator)
        return iterator

    async def __aiter__(self):
        """[Async] Iterate keys in cache including expired items."""
        iterator = self._aiter()
        await anext(iterator)
        return iterator
    
    def __reversed__(self):
        """Reverse iterate keys in cache including expired items."""
        iterator = self._iter(ascending=False)
        next(iterator)
        return iterator

    def length(self) -> int:
        """Return the length of the cache.
        """
        return self.reset('count')
    
    async def alength(self) -> int:
        """[Async] Return the length of the cache.
        """
        return await self.areset('count')

    def __hash__(self):
        """Return the hash of the cache.
        """
        return hash(f'{self.table}:{self.conn_uri}')


    def __enter__(self):
        # Create connection in thread.
        # pylint: disable=unused-variable
        self._init_io_()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        # Close connection in thread.
        # pylint: disable=unused-variable
        self.close()
    
    async def __aenter__(self):
        # Create connection in thread.
        # pylint: disable=unused-variable
        await self._init_aio_()
        return self
    
    async def __aexit__(self, exc_type, exc_val, exc_tb):
        # Close connection in thread.
        # pylint: disable=unused-variable    
        await self.aclose()



from __future__ import annotations

"""
A FastAPI Exception Helper to register errors
"""

import contextlib
import traceback
from fastapi import FastAPI, Request, status
from fastapi.exceptions import HTTPException, RequestValidationError
from fastapi.encoders import jsonable_encoder
from .responses import JSONResponse, PrettyJSONResponse
from typing import Optional, Dict, Any, Union, Type, List, Callable, Set, TYPE_CHECKING
from lzl.logging import logger


FASTAPI_ERRORS: Dict[str, List[Type['FastAPIException']]] = {}
ADDED_ERRORS: Set['FastAPIException'] = set()

def register_error(error: Type['FastAPIException']):
    """
    Registers an error
    """
    global FASTAPI_ERRORS
    if error.module is None: 
        error.module = error.__module__.split('.')[0]
    if error.module not in FASTAPI_ERRORS: FASTAPI_ERRORS[error.module] = []
    FASTAPI_ERRORS[error.module].append(error)


class FastAPIException(HTTPException):
    """
    FastAPI Exception
    """
    module: Optional[str] = None
    base: Optional[str] = None
    concat_detail: Optional[bool] = True
    verbose: Optional[int] = None
    
    kind: Optional[str] = 'fastapi_error'
    status: Optional[int] = 400 # Default status code
    problem: Optional[str] = None
    solution: Optional[str] = None

    log_error: Optional[bool] = False # Always log the error message

    def __init_subclass__(cls) -> None:
        """
        Registers the error
        """
        register_error(cls)
        return super().__init_subclass__()
    

    def __init__(
        self, 
        detail: str = None, 
        error: Optional[Exception] = None, 
        status_code: Optional[int] = None, 
        headers: Optional[Dict[str, str]] = None,
        **kwargs
    ):
        """
        Constructor
        """
        self.constr_msg(detail = detail, error = error, **kwargs)
        self.conf(**kwargs)
        self.headers = headers
        self.status_code = status_code or self.status
        self.kwargs = kwargs
        if self.log_error: self.display(self.detail, self.status_code)

    def constr_msg(self, detail: str = None, error: Optional[Exception] = None, **kwargs):
        """
        Constructor
        """
        message = ""
        if (not detail or self.concat_detail) and self.base:
            message += f"{self.base}: " or ''
        if detail: message += detail
        if error and self.verbose is not None:
            if self.verbose >= 1: message += f"\nError: {error}"
            elif self.verbose <= 5: message += f"\nError: {error}\nTraceback: {traceback.format_exc()}"
        with contextlib.suppress(Exception):
            if '{' in message: message = message.format(**kwargs)
        self.detail = message
        with contextlib.suppress(Exception):
            if '{' in self.problem: self.problem = self.problem.format(**kwargs)
            if '{' in self.solution: self.solution = self.solution.format(**kwargs)
        

    def conf(self, **kwargs):
        """
        Configures the error
        """
        for kw in {'verbose', 'concat_detail', 'base', 'kind', 'problem', 'solution'}:
            if kw in kwargs: setattr(self, kw, kwargs.get(kw))
        
    def display(self, message: str, status_code: int):
        """
        Displays the error
        """
        logger.error(f'[{self.__class__.__name__} - {status_code}]: {message}')

    def __str__(self):
        """
        String Representation
        """
        return f'[{self.__class__.__name__} - {self.status_code}]: {self.detail}'
    
    def __repr__(self):
        """
        Representation
        """
        return f'{self.__class__.__name__}(status_code={self.status_code!r}, detail={self.detail!r})'
    

    def json_data(self) -> Dict[str, Union[str, int]]:
        """
        Returns the json data
        """
        return {
            'kind': self.kind,
            'status': 'error',
            'status_code': self.status_code,
            'problem': self.problem,
            'solution': self.solution,
            'detail': self.detail,
            'error': self.__class__.__name__,
        }



async def default_validation_exception_handler(request: Request, exc: RequestValidationError):
    """
    Handle validation errors
    """
    errors = exc.errors()
    log_errors = jsonable_encoder(errors)
    logger.error(f"Validation error for request {request.url}: {log_errors}")
    return JSONResponse(
        status_code = status.HTTP_422_UNPROCESSABLE_ENTITY,
        content = jsonable_encoder({"detail": errors}),
    )


async def default_exception_handler(request: Request, exc: FastAPIException):
    """
    Default Exception Handler
    """
    logger.error(f"Error: {exc}")
    return PrettyJSONResponse(
        status_code = exc.status_code,
        content = exc.json_data(),
    )


def add_exception_handlers_to_app(
    app: 'FastAPI',
    func: Optional[Callable] = None,
    module: Optional[str] = None,
    verbose: Optional[bool] = False,
    detailed: Optional[bool] = False,
):
    """
    Adds the exception handlers to the app
    """
    global ADDED_ERRORS
    if module is None: module = list(FASTAPI_ERRORS.keys())[0]
    seen = set()
    func = func or default_exception_handler
    for error in FASTAPI_ERRORS.get(module, []):
        # if error.__name__ in seen: continue
        if error in ADDED_ERRORS: continue
        if verbose and detailed: logger.info(f'Adding Exception Handler: |g|{error.__name__}|e|', colored = True, prefix = f'{module}')
        app.add_exception_handler(error, func)
        ADDED_ERRORS.add(error)
        seen.add(error.__name__)
    if verbose: logger.info(f'Added |g|{len(seen)}|e| Exception Handlers to |g|{module}|e|', colored = True)

def add_all_exceptions_to_app(
    app: 'FastAPI',
    func: Optional[Callable] = None,
    verbose: Optional[bool] = False,
    detailed: Optional[bool] = False,
):
    """
    Adds all the exception handlers to the app
    """
    for module in FASTAPI_ERRORS:
        add_exception_handlers_to_app(app, func = func, module = module, verbose = verbose, detailed = detailed)


from __future__ import annotations

"""
Extra Types
"""


from lzo.types import BaseModel, Field, eproperty, model_validator
from typing import Optional, List, Dict, Any, Union, Type, TYPE_CHECKING

class UserInfo(BaseModel):
    """
    User Info
    """

    id: Optional[str] = None
    provided_id: Optional[str] = None
    preferred_email: Optional[str] = None
    username: Optional[str] = None
    last_name: Optional[str] = None
    first_name: Optional[str] = None
    is_suspended: Optional[bool] = None
    picture: Optional[str] = None
    total_sign_ins: Optional[int] = None
    last_signed_in: Optional[str] = None
    created_on: Optional[str] = None

    organizations: Optional[List[str]] = None
    identities: Optional[List[Dict[str, str]]] = None


from __future__ import annotations

from typing import Any, Dict, Optional, Union, Type, Tuple, TYPE_CHECKING
from lzl.load import lazy_import
from .base import BaseSerializer, ObjectValue, SchemaType, SerializableObject, BaseModel, logger, ThreadPool, ModuleType
from .utils import serialize_object
from .defaults import default_json, JsonLibT

class JsonSerializer(BaseSerializer):

    name: Optional[str] = "json"
    encoding: Optional[str] = "utf-8"
    jsonlib: JsonLibT = default_json
    disable_object_serialization: Optional[bool] = False
    disable_nested_values: Optional[bool] = None
    allow_failed_import: Optional[bool] = False

    def __init__(
        self, 
        jsonlib: Optional[Union[str, Any]] = None,
        compression: Optional[str] = None,
        compression_level: int | None = None, 
        encoding: str | None = None, 
        serialization_obj: Optional[Type[BaseModel]] = None,
        serialization_obj_kwargs: Optional[Dict[str, Any]] = None,
        disable_object_serialization: Optional[bool] = None,
        disable_nested_values: Optional[bool] = None,
        verbosity: Optional[int] = None,
        **kwargs
    ):
        super().__init__(compression = compression, compression_level = compression_level, encoding = encoding, **kwargs)
        self.serialization_obj = serialization_obj
        self.serialization_obj_kwargs = serialization_obj_kwargs or {}
        self.serialization_schemas: Dict[str, Type[BaseModel]] = {}
        if disable_object_serialization is not None:
            self.disable_object_serialization = disable_object_serialization
        if disable_nested_values is not None:
            self.disable_nested_values = disable_nested_values
        if jsonlib is not None:
            if isinstance(jsonlib, str):
                jsonlib = lazy_import(jsonlib, is_module=True)
            assert hasattr(jsonlib, "dumps") and hasattr(jsonlib, "loads"), f"Invalid JSON Library: {jsonlib}"
            self.jsonlib = jsonlib
        self.verbosity = verbosity
        self.jsonlib_name: str = self.jsonlib.__name__
        if 'bindings' in self.jsonlib_name.lower():
            self.jsonlib_name = self.jsonlib_name.rsplit('_', 1)[-1]

    @classmethod
    def set_default_lib(cls, lib: Union[str, JsonLibT, ModuleType]) -> None:
        """
        Sets the default JSON library
        """
        global default_json
        if isinstance(lib, str):
            lib = lazy_import(lib, is_module=True)
        assert hasattr(lib, "dumps") and hasattr(lib, "loads"), f"Invalid JSON Library: {lib}"
        cls.jsonlib = lib
        default_json = lib

    @property
    def _is_verbose(self) -> bool:
        """
        Returns whether the serializer is verbose
        """
        return self.verbosity is None or self.verbosity >= 1

    @property
    def _is_silenced(self) -> bool:
        """
        Returns whether the serializer is verbose
        """
        return self.verbosity and self.verbosity < 0
        
    def serialize_obj(self, obj: SerializableObject, mode: Optional[SerMode] = None, **kwargs) -> Union[str, bytes]:
        """
        Serializes the object
        """
        mode = mode or self.ser_mode
        if 'disable_nested_values' not in kwargs and self.disable_nested_values is not None:
            kwargs['disable_nested_values'] = self.disable_nested_values
        return serialize_object(obj, mode = mode, **kwargs)
    

    def encode_value(self, value: Union[Any, SchemaType], **kwargs) -> str:
        """
        Encode the value with the JSON Library
        """
        try:
            value_dict = self.serialize_obj(value, **kwargs, **self.serialization_obj_kwargs)
            encoded = self.jsonlib.dumps(value_dict, **kwargs)
            return self.coerce_output_value(encoded)

        except Exception as e:
            if not self._is_silenced: logger.trace(f'Error Encoding Value: |r|({type(value)})|e| {str(value)[:1000]}', e, colored = True)
        try:
            encoded = self.jsonlib.dumps(value, **kwargs)
            return self.coerce_output_value(encoded)
        except Exception as e:
            if not self._is_silenced: logger.info(f'Error Encoding Value: |r|({type(value)}) {e}|e| {str(value)[:1000]}', colored = True, prefix = self.jsonlib_name)
            if self.raise_errors: raise e
        return None


    def decode(self, value: Union[str, bytes], schema_map: Optional[Dict[str, str]] = None, raise_errors: Optional[bool] = None, **kwargs) -> ObjectValue:
        """
        Decodes the value
        """
        try:
            decompressed_value = self.decompress_value(value, **kwargs)
            if decompressed_value is not None:
                value = decompressed_value
        except Exception as e:
            if not self._is_silenced: logger.info(f'Error Decompressing Value: |r|({type(value)}) {e}|e| {str(value)[:100]}', colored = True, prefix = self.jsonlib_name)
            if raise_errors or self.raise_errors: raise ValueError(f"[{self.name}] Error in Decompression: {str(value)[:100]}") from e
            # return self.decode_value(value, **kwargs)
        return self.decode_value(value, schema_map = schema_map, raise_errors = raise_errors, **kwargs)
    
    
    def decode_value(self, value: str, schema_map: Optional[Dict[str, str]] = None, raise_errors: Optional[bool] = None, **kwargs) -> Union[SchemaType, Dict, Any]:
        """
        Decode the value with the JSON Library
        """
        if value is None: return None
        if isinstance(value, (str, bytes)):
            try:
                # value = self.check_encoded_value(value)
                value = self.jsonlib.loads(value, **kwargs)
            except Exception as e:
                if isinstance(value, str) and 'Exception' in value or 'Traceback (most recent call last):' in value:
                    return value
                str_value = str(value)
                if not schema_map: str_value = str_value[:1000]
                if self._is_verbose: logger.info(f'Error JSON Decoding Value: |r|({type(value)}) {e}|e| {str_value}', colored = True, prefix = self.jsonlib_name)
                if raise_errors or self.raise_errors: raise e
        try:
            return self.deserialize_obj(value, schema_map = schema_map, allow_failed_import = self.allow_failed_import)
        except Exception as e:
            str_value = str(value)
            if not schema_map: str_value = str_value[:1000]
            if not self._is_silenced: logger.trace(f'Error Deserializing Object: ({type(value)}) {str_value}', e, prefix = self.jsonlib_name)
            if raise_errors or self.raise_errors: raise e
        return None


    async def adecode(self, value: Union[str, bytes], schema_map: Optional[Dict[str, str]] = None, raise_errors: Optional[bool] = None, **kwargs) -> ObjectValue:
        """
        Decodes the value asynchronously
        """
        return await ThreadPool.arun(self.decode, value, schema_map = schema_map, raise_errors = raise_errors, **kwargs)

        
    if TYPE_CHECKING:
        def dumps(
            self, 
            value: ObjectValue, 
            skipkeys: bool = False, 
            ensure_ascii: bool = True, 
            check_circular: bool = True,
            allow_nan: bool = True, 
            cls: Optional[Any] = None, 
            indent: Optional[int] = None, 
            separators: Optional[Tuple[str, str]] = None,
            default: Optional[Any] = None, 
            sort_keys: bool = False,
            **kwargs
        ) -> Union[str, bytes]:
            """Serialize ``obj`` to a JSON formatted ``str``.

            If ``skipkeys`` is true then ``dict`` keys that are not basic types
            (``str``, ``int``, ``float``, ``bool``, ``None``) will be skipped
            instead of raising a ``TypeError``.

            If ``ensure_ascii`` is false, then the return value can contain non-ASCII
            characters if they appear in strings contained in ``obj``. Otherwise, all
            such characters are escaped in JSON strings.

            If ``check_circular`` is false, then the circular reference check
            for container types will be skipped and a circular reference will
            result in an ``RecursionError`` (or worse).

            If ``allow_nan`` is false, then it will be a ``ValueError`` to
            serialize out of range ``float`` values (``nan``, ``inf``, ``-inf``) in
            strict compliance of the JSON specification, instead of using the
            JavaScript equivalents (``NaN``, ``Infinity``, ``-Infinity``).

            If ``indent`` is a non-negative integer, then JSON array elements and
            object members will be pretty-printed with that indent level. An indent
            level of 0 will only insert newlines. ``None`` is the most compact
            representation.

            If specified, ``separators`` should be an ``(item_separator, key_separator)``
            tuple.  The default is ``(', ', ': ')`` if *indent* is ``None`` and
            ``(',', ': ')`` otherwise.  To get the most compact JSON representation,
            you should specify ``(',', ':')`` to eliminate whitespace.

            ``default(obj)`` is a function that should return a serializable version
            of obj or raise TypeError. The default simply raises TypeError.

            If *sort_keys* is true (default: ``False``), then the output of
            dictionaries will be sorted by key.

            To use a custom ``JSONEncoder`` subclass (e.g. one that overrides the
            ``.default()`` method to serialize additional types), specify it with
            the ``cls`` kwarg; otherwise ``JSONEncoder`` is used.

            """
            ...

    

    







from __future__ import annotations

import typing as t
from .base import RegistryItem
from .main import TemporalRegistry, registry

if t.TYPE_CHECKING:
    from ..mixins import (
        BaseTemporalMixin, TemporalWorkflowMixin, TemporalActivityMixin,
        TemporalMixinT, TemporalWorkflowT, TemporalActivityT, 
        MixinKinds
    )
    from temporalio.types import CallableType, ClassType



from __future__ import annotations

"""
FastAPI Responses
"""


import yaml
from fastapi.responses import (
    JSONResponse as BaseJSONResponse,
    Response,
)
from .utils import json_serializer as serializer
from typing import Any, Optional


class JSONResponse(BaseJSONResponse):

    indention: Optional[int] = None

    def render(self, content: Any) -> bytes:
        """
        Renders the content
        """
        return serializer.dumps(
            content,
            ensure_ascii = False,
            allow_nan = False,
            indent = self.indention,
            separators=(",", ":"),
        ).encode("utf-8")


class PrettyJSONResponse(JSONResponse):
    indention = 4


class YAMLResponse(Response):

    content_type: str = "application/yaml"
    indention: int = 2

    def render(self, content: Any) -> bytes:
        """
        Renders the content
        """
        return yaml.safe_dump(
            content,
            default_flow_style = False,
            indent = self.indention,
        ).encode("utf-8")


from __future__ import annotations

import fastembed.embedding
import fastembed.rerank.cross_encoder.onnx_text_cross_encoder

"""
Qdrant Configuration
"""

import os
import warnings
import typing as t

warnings.filterwarnings("ignore", message = "Api key is used with an insecure connection.", category = UserWarning)

# warnings.filterwarnings("ignore", message = "Api key is used with an insecure connection.", module = "qdrant_client")
import pathlib
from lzl import load
from lzl.proxied import ProxyObject
from lzo.types import BaseModel, BaseSettings, AppEnv, eproperty
from pydantic import model_validator, Field, PrivateAttr
from lzl.logging import logger, null_logger, Logger

if t.TYPE_CHECKING:
    from lzl.io import TemporaryData
    import fastembed
    import fastembed.text.clip_embedding
    import fastembed.text.multitask_embedding
    import fastembed.text.onnx_embedding
    import fastembed.text.pooled_embedding
    import fastembed.text.pooled_normalized_embedding
    import fastembed.rerank.cross_encoder
    import fastembed.sparse.splade_pp
    # import fastembed.rerank.cross_encoder.onnx_text_cross_encoder
    from .full_client import QdrantClient
    
else:
    fastembed = load.LazyLoad('fastembed', install_missing = True)


"""
Config File Schema:

pooled_normalized_models:
- model: Alibaba-NLP/gte-modernbert-base
  dim: 768
  description: Text embeddings, Unimodal (text), English, 8192 input tokens truncation, Prefixes for queries/documents: not necessary, 2024 year.
  license: apache-2.0
  size_in_GB: 0.59
  sources:
    hf: Alibaba-NLP/gte-modernbert-base
  model_file: onnx/model.onnx
"""


lib_path = pathlib.Path(__file__).parent


class QdrantSharedConfig(BaseModel):
    """
    The shared config for Qdrant
    """
    location: t.Optional[str] = None
    url: t.Optional[str] = None
    port: t.Optional[int] = None
    grpc_port: t.Optional[int] = None
    prefer_grpc: t.Optional[bool] = None
    https: t.Optional[bool] = None
    api_key: t.Optional[str] = None
    prefix: t.Optional[str] = None
    timeout: t.Optional[int] = None
    host: t.Optional[str] = None
    path: t.Optional[str] = None
    
    force_disable_check_same_thread: t.Optional[bool] = None
    auth_token_provider: t.Optional[
        t.Union[t.Callable[[], str], t.Callable[[], t.Awaitable[str]]]
    ] = None
    cloud_inference: t.Optional[bool] = None
    check_compatibility: t.Optional[bool] = None
    kwargs: t.Optional[t.Dict[str, t.Any]] = Field(None, exclude = True)

    set_model: t.Optional[str] = Field(None, exclude = True)
    set_model_config: t.Optional[t.Dict[str, t.Any]] = Field(default_factory=dict, exclude = True)
    set_sparse_model: t.Optional[str] = Field(None, exclude = True)
    set_sparse_model_config: t.Optional[t.Dict[str, t.Any]] = Field(default_factory=dict, exclude = True)
    retries: t.Optional[int] = Field(None, exclude = True)


    @classmethod
    def build(
        cls, 
        # s: 'QdrantClientSettings',
        c: 'QdrantClient',

        # These are shard with the config
        url: t.Optional[str] = None,
        port: t.Optional[int] = None,
        grpc_port: t.Optional[int] = None,
        prefer_grpc: t.Optional[bool] = None,
        https: t.Optional[bool] = None,
        api_key: t.Optional[str] = None,
        prefix: t.Optional[str] = None,
        timeout: t.Optional[int] = None,
        host: t.Optional[str] = None,
        path: t.Optional[str] = None,

        # These are from the client
        location: t.Optional[str] = None,
        force_disable_check_same_thread: t.Optional[bool] = None,
        grpc_options: t.Optional[t.Dict[str, t.Any]] = None,
        auth_token_provider: t.Optional[
            t.Union[t.Callable[[], str], t.Callable[[], t.Awaitable[str]]]
        ] = None,
        cloud_inference: t.Optional[bool] = None,
        check_compatibility: t.Optional[bool] = None,
        set_model: t.Optional[str] = None,
        set_sparse_model: t.Optional[str] = None,
        retries: t.Optional[int] = None,
        **kwargs: t.Any,
    ) -> 'QdrantSharedConfig':
        """
        Constructs a shared client config
        """
        new = {
            'url': url if url is not None else c.settings.url,
            'port': port or c.settings.port,
            'grpc_port': grpc_port or c.settings.grpc_port,
            'prefer_grpc': prefer_grpc if prefer_grpc is not None else c.settings.prefer_grpc,
            'https': https if https is not None else c.settings.https,
            'api_key': api_key if api_key is not None else c.settings.api_key,
            'prefix': prefix if prefix is not None else c.settings.prefix,
            'timeout': timeout if timeout is not None else c.settings.timeout,
            'host': host if host is not None else c.settings.host,
            'path': path if path is not None else c.settings.path,
            'location': location,
            'force_disable_check_same_thread': force_disable_check_same_thread,
            'grpc_options': grpc_options,
            'auth_token_provider': auth_token_provider,
            'cloud_inference': cloud_inference,
            'check_compatibility': check_compatibility,
            'set_model': set_model,
            'set_sparse_model': set_sparse_model,
            'retries': retries if retries is not None else c.settings.retries,
            'kwargs': kwargs,
        }
        new = {k:v for k,v in new.items() if v is not None}
        return cls.model_validate(new, context = {'source': 'build', 'client': c})

class QdrantClientSettings(BaseSettings):
    """
    The Qdrant Client Settings
    """

    url: t.Optional[str] = None
    port: t.Optional[int] = 6333
    grpc_port: t.Optional[int] = 6334
    prefer_grpc: t.Optional[bool] = False
    https: t.Optional[bool] = None
    api_key: t.Optional[str] = None
    prefix: t.Optional[str] = None
    timeout: t.Optional[int] = None
    host: t.Optional[str] = None
    path: t.Optional[str] = None

    # Extra
    fastembed_config: t.Optional[str] = None
    set_model: t.Optional[str] = None
    set_sparse_model: t.Optional[str] = None
    retries: t.Optional[int] = None


    class Config(BaseSettings.Config):
        env_prefix = "QDRANT_"

    @eproperty
    def temp_data(self) -> 'TemporaryData':
        """
        Returns the temporary data
        """
        from lzl.io.persistence import TemporaryData
        return TemporaryData(lib_path.joinpath('data'))

    @eproperty
    def added_fastembed_models(self) -> t.Set[str]:
        """
        Returns the already added fastembed models
        """
        return set()
        # return self._extra.get('added_fastembed_models', set())

    
    def add_fastembed_models_v1(
        self,
        
        # embedding
        clip_models: t.Optional[t.List[t.Dict[str, t.Union[str, int, t.Dict[str, str]]]]] = None,
        multitask_models: t.Optional[t.List[t.Dict[str, t.Union[str, int, t.Dict[str, str]]]]] = None,
        pooled_models: t.Optional[t.List[t.Dict[str, t.Union[str, int, t.Dict[str, str]]]]] = None,
        pooled_normalized_models: t.Optional[t.List[t.Dict[str, t.Union[str, int, t.Dict[str, str]]]]] = None,
        onnx_models: t.Optional[t.List[t.Dict[str, t.Union[str, int, t.Dict[str, str]]]]] = None,

        # added for reranking
        cross_encoder_models: t.Optional[t.List[t.Dict[str, t.Union[str, int, t.Dict[str, str]]]]] = None,
        
        # added for sparse models
        sparse_splade_pp_models: t.Optional[t.List[t.Dict[str, t.Union[str, int, t.Dict[str, str]]]]] = None,

        verbose: t.Optional[bool] = False,
        **kwargs: t.Any,
    ) -> None:
        """
        [v1] Adds the specified models to the list of added models

        This method is for adding models before v0.6.0
        """
        # Embedding
        if clip_models:
            for conf in clip_models:
                if conf['model'] not in self.added_fastembed_models:
                    if verbose: logger.info(f'Adding Model |g|{conf["model"]}|e| to `clip_models`', colored = True)
                    fastembed.text.clip_embedding.supported_clip_models.append(conf)
                    self.added_fastembed_models.add(conf['model'])

        if multitask_models:
            try:
                for conf in multitask_models:
                    if conf['model'] not in self.added_fastembed_models:
                        if verbose: logger.info(f'Adding Model |g|{conf["model"]}|e| to `multitask_models`', colored = True)
                        fastembed.text.multitask_embedding.supported_multitask_models.append(conf)
                        self.added_fastembed_models.add(conf['model'])
            except Exception as e:
                logger.trace('Unable to add multi-task models', e)

        if pooled_models:
            for conf in pooled_models:
                if conf['model'] not in self.added_fastembed_models:
                    if verbose: logger.info(f'Adding Model |g|{conf["model"]}|e| to `pooled_models`', colored = True)
                    fastembed.text.pooled_embedding.supported_pooled_models.append(conf)
                    self.added_fastembed_models.add(conf['model'])

        if pooled_normalized_models:
            for conf in pooled_normalized_models:
                if conf['model'] not in self.added_fastembed_models:
                    if verbose: logger.info(f'Adding Model |g|{conf["model"]}|e| to `pooled_normalized_models`', colored = True)
                    fastembed.text.pooled_normalized_embedding.supported_pooled_normalized_models.append(conf)
                    self.added_fastembed_models.add(conf['model'])

        if onnx_models:
            for conf in onnx_models:
                if conf['model'] not in self.added_fastembed_models:
                    if verbose: logger.info(f'Adding Model |g|{conf["model"]}|e| to `onnx_models`', colored = True)
                    fastembed.text.pooled_normalized_embedding.supported_pooled_normalized_models.append(conf)
                    self.added_fastembed_models.add(conf['model'])

        # Rerankers
        if cross_encoder_models:
            for conf in cross_encoder_models:
                if conf['model'] not in self.added_fastembed_models:
                    if verbose: logger.info(f'Adding Model |g|{conf["model"]}|e| to `cross_encoder_models`', colored = True)
                    fastembed.rerank.cross_encoder.onnx_text_cross_encoder.supported_onnx_models.append(conf)
                    self.added_fastembed_models.add(conf['model'])
        
        # Sparse Models
        if sparse_splade_pp_models:
            for conf in sparse_splade_pp_models:
                if conf['model'] not in self.added_fastembed_models:
                    if verbose: logger.info(f'Adding Model |g|{conf["model"]}|e| to `sparse_splade_pp_models`', colored = True)
                    fastembed.sparse.splade_pp.supported_splade_models.append(conf)
                    self.added_fastembed_models.add(conf['model'])

    def add_fastembed_models(
        self,
        
        # embedding
        clip_models: t.Optional[t.List[t.Dict[str, t.Union[str, int, t.Dict[str, str]]]]] = None,
        multitask_models: t.Optional[t.List[t.Dict[str, t.Union[str, int, t.Dict[str, str]]]]] = None,
        pooled_models: t.Optional[t.List[t.Dict[str, t.Union[str, int, t.Dict[str, str]]]]] = None,
        pooled_normalized_models: t.Optional[t.List[t.Dict[str, t.Union[str, int, t.Dict[str, str]]]]] = None,
        onnx_models: t.Optional[t.List[t.Dict[str, t.Union[str, int, t.Dict[str, str]]]]] = None,

        # added for reranking
        cross_encoder_models: t.Optional[t.List[t.Dict[str, t.Union[str, int, t.Dict[str, str]]]]] = None,
        
        # added for sparse models
        sparse_splade_pp_models: t.Optional[t.List[t.Dict[str, t.Union[str, int, t.Dict[str, str]]]]] = None,

        verbose: t.Optional[bool] = False,
        **kwargs: t.Any,
    ) -> None:
        """
        Adds the specified models to the list of added models
        """
        try:
            from fastembed.common.model_description import PoolingType, BaseModelDescription, ModelSource, DenseModelDescription, SparseModelDescription
        except ImportError:
            return self.add_fastembed_models_v1(
                clip_models=clip_models,
                multitask_models=multitask_models,
                pooled_models=pooled_models,
                pooled_normalized_models=pooled_normalized_models,
                onnx_models=onnx_models,
                cross_encoder_models=cross_encoder_models,
                sparse_splade_pp_models=sparse_splade_pp_models,
                verbose=verbose,
                **kwargs,
            )
        
        # > 0.6.0
        # CLIP Models 
        if clip_models:
            for conf in clip_models:
                if conf['model'] not in self.added_fastembed_models:
                    if verbose: logger.info(f'Adding Model |g|{conf["model"]}|e| to `clip_models`', colored = True)
                    sources = ModelSource(**conf.pop('sources'))
                    fastembed.text.clip_embedding.supported_clip_models.append(
                        DenseModelDescription(
                            sources = sources,
                            **conf
                        )
                    )
                    self.added_fastembed_models.add(conf['model'])
    
        if multitask_models:
            try:
                for conf in multitask_models:
                    if conf['model'] not in self.added_fastembed_models:
                        if verbose: logger.info(f'Adding Model |g|{conf["model"]}|e| to `multitask_models`', colored = True)
                        sources = ModelSource(**conf.pop('sources'))
                        fastembed.text.multitask_embedding.supported_multitask_models.append(
                            DenseModelDescription(
                                sources = sources,
                                **conf
                            )
                        )
                        self.added_fastembed_models.add(conf['model'])
            except Exception as e:
                logger.trace('Unable to add multi-task models', e)

        if pooled_models:
            for conf in pooled_models:
                if conf['model'] not in self.added_fastembed_models:
                    if verbose: logger.info(f'Adding Model |g|{conf["model"]}|e| to `pooled_models`', colored = True)
                    sources = ModelSource(**conf.pop('sources'))
                    fastembed.text.pooled_embedding.supported_pooled_models.append(
                        DenseModelDescription(
                            sources = sources,
                            **conf
                        )
                    )
                    # fastembed.text.pooled_embedding.supported_pooled_models.append(conf)
                    self.added_fastembed_models.add(conf['model'])

        if pooled_normalized_models:
            for conf in pooled_normalized_models:
                if conf['model'] not in self.added_fastembed_models:
                    if verbose: logger.info(f'Adding Model |g|{conf["model"]}|e| to `pooled_normalized_models`', colored = True)
                    sources = ModelSource(**conf.pop('sources'))
                    fastembed.text.pooled_normalized_embedding.supported_pooled_normalized_models.append(
                        DenseModelDescription(
                            sources = sources,
                            **conf
                        )
                    )
                    self.added_fastembed_models.add(conf['model'])

        if onnx_models:
            for conf in onnx_models:
                if conf['model'] not in self.added_fastembed_models:
                    if verbose: logger.info(f'Adding Model |g|{conf["model"]}|e| to `onnx_models`', colored = True)
                    sources = ModelSource(**conf.pop('sources'))
                    fastembed.text.onnx_embedding.supported_onnx_models.append(
                        DenseModelDescription(
                            sources = sources,
                            **conf
                        )
                    )
                    self.added_fastembed_models.add(conf['model'])

        # Rerankers
        if cross_encoder_models:
            for conf in cross_encoder_models:
                if conf['model'] not in self.added_fastembed_models:
                    if verbose: logger.info(f'Adding Model |g|{conf["model"]}|e| to `cross_encoder_models`', colored = True)
                    sources = ModelSource(**conf.pop('sources'))
                    fastembed.rerank.cross_encoder.onnx_text_cross_encoder.supported_onnx_models.append(
                        BaseModelDescription(
                            sources = sources,
                            **conf
                        )
                    )
                    self.added_fastembed_models.add(conf['model'])
        
        # Sparse Models
        if sparse_splade_pp_models:
            for conf in sparse_splade_pp_models:
                if conf['model'] not in self.added_fastembed_models:
                    if verbose: logger.info(f'Adding Model |g|{conf["model"]}|e| to `sparse_splade_pp_models`', colored = True)
                    sources = ModelSource(**conf.pop('sources'))
                    fastembed.sparse.splade_pp.supported_splade_models.append(
                        SparseModelDescription(
                            sources = sources,
                            **conf
                        )
                    )
                    self.added_fastembed_models.add(conf['model'])

        
    
    def _load_fastembed_configs(self, config: t.Optional[str]):
        """
        Loads the fastembed configs from a file
        """
        import yaml
        from lzl.io import File
        config_file = File(config)
        if not config_file.exists(): raise ValueError(f'Config File cannot be loaded from: `{config}`')
        if not self.temp_data.has_logged(f'config_file:load:{config_file.as_posix()}'):
            logger.info(f'Loading from Config File: `{config_file}`')
        
        config_data: t.Dict[str, t.List[t.Dict[str, t.Union[str, int, t.Dict[str, str]]]]] = yaml.safe_load(config_file.read_text())
        self.add_fastembed_models(**config_data)


    @model_validator(mode = 'after')
    def validate_config(self):
        """
        Validates the config
        """
        if self.fastembed_config: self._load_fastembed_configs(self.fastembed_config)
        return self
    

settings: QdrantClientSettings = ProxyObject(
    QdrantClientSettings
)


from __future__ import annotations

import io
import anyio
import asyncio
import contextlib
import errno
from typing import Optional, Dict, Any, Union, List, TYPE_CHECKING

with contextlib.suppress(ImportError):
    import s3fs
    from fsspec.asyn import AbstractAsyncStreamedFile, sync, get_loop
    from s3fs.core import _inner_fetch, _fetch_range as _sync_fetch_range
    from s3fs.utils import FileExpired

if TYPE_CHECKING:
    from .filesys import R2FileSystem

from .utils import _log


"""
Base R2 File
"""

class R2File(s3fs.S3File):
    fs: 'R2FileSystem' = None
    buffer: io.BytesIO = None

    def _call_s3(self, method, *kwarglist, **kwargs):
        """
        Filter out ACL for methods that we know will fail
        """
        # if method in ["create_multipart_upload", "put_object", "put_object_acl"]:
        acl = kwargs.pop("ACL", None)
        _log(f'calling method: {method}')
        return self.fs.call_s3(method, self.s3_additional_kwargs, *kwarglist, **kwargs)


    def _upload_chunk(self, final=False):
        bucket, key, _ = self.fs.split_path(self.path)
        _log(
            f"Upload for {self}, final={final}, loc={self.loc}, buffer loc={self.buffer.tell()}"
        )
        if (
            self.autocommit
            and not self.append_block
            and final
            and self.tell() < self.blocksize
        ):
            # only happens when closing small file, use on-shot PUT
            data1 = False
        else:
            self.buffer.seek(0)
            (data0, data1) = (None, self.buffer.read(self.blocksize))

        while data1:
            (data0, data1) = (data1, self.buffer.read(self.blocksize))
            data1_size = len(data1)

            if 0 < data1_size < self.blocksize:
                remainder = data0 + data1
                remainder_size = self.blocksize + data1_size

                if remainder_size <= self.part_max:
                    (data0, data1) = (remainder, None)
                else:
                    partition = remainder_size // 2
                    (data0, data1) = (remainder[:partition], remainder[partition:])

            part = len(self.parts) + 1
            _log(f"Upload chunk {self}, {part}, {self.blocksize}/{data1_size}")

            out = self._call_s3(
                "upload_part",
                Bucket=bucket,
                PartNumber=part,
                UploadId=self.mpu["UploadId"],
                Body=data0,
                Key=key,
            )

            self.parts.append({"PartNumber": part, "ETag": out["ETag"]})

        if self.autocommit and final:
            self.commit()
        return not final
    

    def commit(self):
        _log(f"Commit {self}")
        if self.tell() == 0:
            if self.buffer is not None:
                _log(f"Empty file committed {self}")
                self._abort_mpu()
                write_result = self.fs.touch(self.path)
        elif not self.parts:
            if self.buffer is None:
                raise RuntimeError

            _log(f"One-shot upload of {self}: {self.key}")
            self.buffer.seek(0)
            data = self.buffer.read()
            write_result = self._call_s3(
                "put_object",
                Key=self.key,
                Bucket=self.bucket,
                Body=data,
                ACL=self.acl,
                **self.kwargs,
            )
        else:
            part_info = {"Parts": self.parts}
            _log(f"Complete multi-part upload for {self}: {self.key} {part_info}")
            try:
                write_result = self._call_s3(
                    "complete_multipart_upload",
                    Bucket=self.bucket,
                    Key=self.key,
                    UploadId=self.mpu["UploadId"],
                    # Action="mpu-complete",
                    # Parts=part_info["Parts"],
                    MultipartUpload=part_info,

                )
                # self.mpu.complete()
            except Exception as e:
                self._abort_mpu()
                if 'All non-trailing' in str(e):
                    _log(f"Attempting to upload large file {self}", verbose = True)
                    try:
                        self._large_upload()
                    except Exception as e2:
                        _log(f"Failed to upload large file {self}", verbose = True)
                        raise e2
                raise e

        if self.fs.version_aware: self.version_id = write_result.get("VersionId")
        # complex cache invalidation, since file's appearance can cause several
        # directories
        self.buffer = None
        parts = self.path.split("/")
        path = parts[0]
        for p in parts[1:]:
            if path in self.fs.dircache and not [
                True for f in self.fs.dircache[path] if f["name"] == f"{path}/{p}"
            ]: self.fs.invalidate_cache(path)
            path = f"{path}/{p}"

    def write(self, data: Union[bytes, bytearray, memoryview]):
        """
        Write data to buffer.

        Buffer only sent on flush() or if buffer is greater than
        or equal to blocksize.

        Parameters
        ----------
        data: bytes
            Set of bytes to be written.
        """
        if self.mode not in {"wb", "ab"}: raise ValueError("File not in write mode")
        if self.closed: raise ValueError("I/O operation on closed file.")
        if self.forced: raise ValueError("This file has been force-flushed, can only close")
        out = self.buffer.write(data)
        self.loc += out
        if self.buffer.tell() >= self.blocksize: self.flush()
        return out

    def _large_upload(self, callbacks: Optional[List[Any]] = None, **kwargs):
        """
        Handles large file uploads bc multipart upload isnt working.
        """
        bucket, key, _ = self.fs.split_path(self.path)
        _log(f"Large upload for {self}, loc={self.loc}, buffer loc={self.buffer.tell()}")
        self.buffer.seek(0)
        future = self.fs.s3tm.upload(
            self.buffer,
            bucket,
            key,
            subscribers = callbacks
        )
        res = future.result()
        _log(f"Large upload for {self} complete: {res}")
        self.buffer = io.BytesIO()


    def flush(self, force: bool = False):
        """
        Write buffered data to backend store.

        Writes the current buffer, if it is larger than the block-size, or if
        the file is being closed.

        Parameters
        ----------
        force: bool
            When closing, write the last block even if it is smaller than
            blocks are allowed to be. Disallows further writing to this file.
        """

        if self.closed: raise ValueError("Flush on closed file")
        if force and self.forced: raise ValueError("Force flush cannot be called more than once")
        if force: self.forced = True

        # no-op to flush on read-mode
        if self.mode not in {"wb", "ab"}: return

        # Defer write on small block
        if not force and self.buffer.tell() < self.blocksize: return
        
        # if self.buffer.tell() >= self.fs.default_r2_max_size:
        # if self.buffer.tell() >= self.fs.default_r2_block_size:
        #     # Write large file
        #     self._large_upload()
        #     return

        # Write large file
        if self.buffer.tell() >= self.fs.default_r2_large_file_threshold:
            self._large_upload()
            return

        if self.offset is None:
            # Initialize a multipart upload
            self.offset = 0
            try: self._initiate_upload()
            except:  # noqa: E722
                self.closed = True
                raise
        
        if self._upload_chunk(final=force) is not False:
            self.offset += self.buffer.seek(0, 2)
            self.buffer = io.BytesIO()


"""
Async Streamed File
"""

class R2AsyncStreamedFile(R2File, AbstractAsyncStreamedFile):
    fs: 'R2FileSystem' = None

    """
    Note, this shouldn't be called for 
    `rb` mode, since it doesnt work with
    the loops for some reason.
    """

    def __init__(self, fs: 'R2FileSystem', path, mode, encoding: Optional[str] = None, errors: Optional[str]  = None, newline: Optional[str]  = None, buffering: Optional[int]  = None, **kwargs):
        super().__init__(fs, path, mode, **kwargs)
        self.r = None
        self.size = None


    async def _acall_s3(self, method, *kwarglist, **kwargs):
        """
        Filter out ACL for methods that we know will fail
        """
        kwargs.pop("ACL", None)
        _log(f'calling method: {method}')
        return await self.fs._call_s3(method, self.s3_additional_kwargs, *kwarglist, **kwargs)


    async def _initiate_upload(self):
        # only happens when closing small file, use on-shot PUT
        if self.autocommit and not self.append_block and self.tell() < self.blocksize: return
        _log(f"Initiate upload for {self}")
        self.parts = []
        self.mpu = await self._acall_s3(
            "create_multipart_upload",
            Bucket=self.bucket,
            Key=self.key,
            ACL=self.acl,
        )

        if self.append_block:
            # use existing data in key when appending,
            # and block is big enough
            out = await self._acall_s3(
                "upload_part_copy",
                self.s3_additional_kwargs,
                Bucket=self.bucket,
                Key=self.key,
                PartNumber=1,
                UploadId=self.mpu["UploadId"],
                CopySource=self.path,
            )
            self.parts.append({"PartNumber": 1, "ETag": out["CopyPartResult"]["ETag"]})

    async def _upload_chunk(self, final: bool = False):
        bucket, key, _ = self.fs.split_path(self.path)
        _log(f"Upload for {self}, final={final}, loc={self.loc}, buffer loc={self.buffer.tell()}")
        
        # only happens when closing small file, use on-shot PUT
        if (
            self.autocommit
            and not self.append_block
            and final
            and self.tell() < self.blocksize
        ): data1 = False
        else:
            self.buffer.seek(0)
            (data0, data1) = (None, self.buffer.read(self.blocksize))

        while data1:
            (data0, data1) = (data1, self.buffer.read(self.blocksize))
            data1_size = len(data1)

            if 0 < data1_size < self.blocksize:
                remainder = data0 + data1
                remainder_size = self.blocksize + data1_size
                if remainder_size <= self.part_max: (data0, data1) = (remainder, None)
                else:
                    partition = remainder_size // 2
                    (data0, data1) = (remainder[:partition], remainder[partition:])

            part = len(self.parts) + 1
            _log(f"Upload chunk {self}, {part}, {self.blocksize}/{data1_size}")
            out = await self._acall_s3(
                "upload_part",
                Bucket=bucket,
                PartNumber=part,
                UploadId=self.mpu["UploadId"],
                Body=data0,
                Key=key,
            )
            self.parts.append({"PartNumber": part, "ETag": out["ETag"]})
        if self.autocommit and final: await self.commit()
        return not final
    
    async def _large_upload(self, callbacks: Optional[List[Any]] = None, **kwargs):
        """
        Handles large file uploads bc multipart upload isnt working.
        """
        bucket, key, _ = self.fs.split_path(self.path)
        _log(f"Large upload for {self}, loc={self.loc}, buffer loc={self.buffer.tell()}")
        self.buffer.seek(0)
        future = self.fs.s3tm.upload(
            self.buffer,
            bucket,
            key,
            subscribers = callbacks
        )
        res = future.result()
        _log(f"Large upload for {self} complete: {res}")
        self.buffer = io.BytesIO()

    async def commit(self):
        _log(f"Commit {self}")
        if self.tell() == 0:
            if self.buffer is not None:
                _log(f"Empty file committed {self}")
                await self._abort_mpu()
                write_result = await self.fs._touch(self.path)
        elif not self.parts:
            if self.buffer is None: raise RuntimeError
            _log(f"One-shot upload of {self}: {self.key}")
            self.buffer.seek(0)
            data = self.buffer.read()
            write_result = await self._acall_s3(
                "put_object",
                Key=self.key,
                Bucket=self.bucket,
                Body=data,
                ACL=self.acl,
                **self.kwargs,
            )
        else:
            part_info = {"Parts": self.parts}
            _log(f"Complete multi-part upload for {self}: {self.key} {part_info}")
            try:
                write_result = await self._acall_s3(
                    "complete_multipart_upload",
                    Bucket=self.bucket,
                    Key=self.key,
                    UploadId=self.mpu["UploadId"],
                    MultipartUpload=part_info,

                )
            except Exception as e:
                await self._abort_mpu()
                if 'All non-trailing' in str(e):
                    _log(f"Attempting to upload large file {self}", verbose = True)
                    try:
                        await self._large_upload()
                    except Exception as e2:
                        _log(f"Failed to upload large file {self}", verbose = True)
                        raise e2
                raise e

        if self.fs.version_aware: self.version_id = write_result.get("VersionId")
        # complex cache invalidation, since file's appearance can cause several
        # directories
        self.buffer = None
        parts = self.path.split("/")
        path = parts[0]
        for p in parts[1:]:
            if path in self.fs.dircache and not [
                True for f in self.fs.dircache[path] if f["name"] == f"{path}/{p}"
            ]: self.fs.invalidate_cache(path)
            path = f"{path}/{p}"

    
    async def write(self, data):
        """
        Write data to buffer.

        Buffer only sent on flush() or if buffer is greater than
        or equal to blocksize.

        Parameters
        ----------
        data: bytes
            Set of bytes to be written.
        """
        if self.mode not in {"wb", "ab"}: raise ValueError("File not in write mode")
        if self.closed: raise ValueError("I/O operation on closed file.")
        if self.forced: raise ValueError("This file has been force-flushed, can only close")
        out = self.buffer.write(data)
        self.loc += out
        if self.buffer.tell() >= self.blocksize: await self.flush()
        return out
    
    async def read(self, length=-1):
        if self.r is None:
            bucket, key, gen = self.fs.split_path(self.path)
            r = await self.fs._call_s3("get_object", Bucket=bucket, Key=key)
            self.size = int(r["ResponseMetadata"]["HTTPHeaders"]["content-length"])
            self.r = r["Body"]
        out = await self.r.read(length)
        self.loc += len(out)
        return out

    async def flush(self, force: bool = False):
        """
        Write buffered data to backend store.

        Writes the current buffer, if it is larger than the block-size, or if
        the file is being closed.

        Parameters
        ----------
        force: bool
            When closing, write the last block even if it is smaller than
            blocks are allowed to be. Disallows further writing to this file.
        """
        if self.closed: raise ValueError("Flush on closed file")
        if force and self.forced: raise ValueError("Force flush cannot be called more than once")
        if force: self.forced = True

        if self.mode not in {"wb", "ab"}: return
        # Defer write on small block
        if not force and self.buffer.tell() < self.blocksize: return
        if self.offset is None:
            self.offset = 0
            try: await self._initiate_upload()
            except:  # noqa: E722
                self.closed = True
                raise

        if await self._upload_chunk(final=force) is not False:
            self.offset += self.buffer.seek(0, 2)
            self.buffer = io.BytesIO()

    
    async def close(self):
        """Close file

        Finalizes writes, discards cache
        """
        if getattr(self, "_unclosable", False): return
        if self.closed: return
        if self.mode == "rb": self.cache = None
        else:
            if not self.forced: await self.flush(force=True)
            if self.fs is not None:
                self.fs.invalidate_cache(self.path)
                self.fs.invalidate_cache(self.fs._parent(self.path))

        self.closed = True

    async def discard(self):
        await self._abort_mpu()
        self.buffer = None  # file becomes unusable

    async def _abort_mpu(self):
        if self.mpu:
            await self._acall_s3(
                "abort_multipart_upload",
                Bucket=self.bucket,
                Key=self.key,
                UploadId=self.mpu["UploadId"],
            )
            self.mpu = None

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()

    


from __future__ import annotations


# from starlette.authentication import BaseUser
from lzo.types import Literal, BaseModel, field_validator, Field, eproperty, PrivateAttr
from lzl.ext.fast.types.roles import UserRole
from typing import Optional, Dict, Any, List, Union, TYPE_CHECKING

if TYPE_CHECKING:
    from ..providers import ProviderClientT, ProviderTokenT
    from ..manager import OAuth2Manager
    from .token import OAuth2Token, AccessToken, IDToken
    from fastapi import Request
    

AuthMethod = Literal['authentication', 'apikey', 'apikey_user', 'session', 'state', None]


class APIKeyUser(BaseModel):
    """
    A Predefined API Key User
    """
    identity: str
    api_key: str
    role: Union[str, int, UserRole]
    email: Optional[str] = None
    allowed_envs: Optional[List[str]] = Field(default_factory = list)
    allowed_domains: Optional[List[str]] = Field(default_factory = list)
    scopes: Optional[List[str]] = Field(default_factory = list)
    provider: Optional[str] = None
    provider_id: Optional[str] = None

    @field_validator("role")
    def validate_role(cls, v: Union[str, int, UserRole]) -> UserRole:
        """
        Validates the role
        """
        return UserRole.parse_role(v) if isinstance(v, (str, int)) else v

    def has_allowed_env(self, env: str) -> bool:
        """
        Checks if the user has an allowed environment
        """
        if not self.allowed_envs: return True
        return True if '*' in self.allowed_envs else env in self.allowed_envs
    
    def has_allowed_domain(self, domain: str) -> bool:
        """
        Checks if the user has an allowed domain
        """
        if not self.allowed_domains: return True
        if '://' in domain: domain = domain.split('://', 1)[-1]
        return True if '*' in self.allowed_domains else domain in self.allowed_domains
    

    @classmethod
    def parse_string(cls , value: str) -> APIKeyUser:
        """
        Parses the string
        - <identity>:<api_key>
        - <identity>:<api_key>:role=<role>
        - <identity>:<api_key>:role=<role>:email=<email>
        - <identity>:<api_key>:role=<role>:email=<email>:allowed_envs=<allowed_envs>
        """
        parts = value.split(':')
        data = {
            'identity': parts[0],
            'api_key': parts[1],
            'role': UserRole.USER,
        }
        parts = parts[2:]
        for part in parts:
            key, value = part.split('=', 1)
            if key == 'role':
                data['role'] = UserRole.parse_role(value)
                continue
            if key == 'email':
                data['email'] = value
                continue
            if key == 'provider':
                data['provider'] = value
                continue
            if key in {
                'scopes', 'allowed_envs', 'allowed_apps', 'allowed_domains'
            }:
                data[key] = value.split(',')
                continue
        return cls(**data)

if TYPE_CHECKING:
    TokenT = Union[AccessToken, IDToken, OAuth2Token, APIKeyUser, ProviderTokenT]

class UserMixin(BaseModel):
    """
    User Mixin
    """
    method: Optional[AuthMethod] = None
    if TYPE_CHECKING:
        token: Optional['TokenT'] = None
    else:
        token: Optional[Any] = None
    
    # apikey: Optional[str] = None
    data: Optional[Union[Dict[str, Any], Any]] = None
    
    is_admin: Optional[bool] = False
    is_staff: Optional[bool] = False
    is_service_account: Optional[bool] = False

    roles: Optional[List[Union[str, UserRole, Dict[str, str]]]] = None
    provider: Optional[str] = None

    # _extra: Dict[str, Any] = {}

    @eproperty
    def apikey(self) -> Optional[str]:
        """
        Returns the API Key
        """
        return self._extra.get('apikey')

    @property
    def is_authenticated(self) -> bool:
        """
        Returns True if the User is authenticated
        """
        return self.method is not None
    
    @property
    def display_name(self) -> str:
        """
        Returns the User's Display Name
        """
        if not self.is_authenticated: return ""
    
    @property
    def identity(self) -> Optional[str]:
        """
        Returns the Identity
        """
        return self.token.identity if self.is_authenticated else None

    @property
    def provider_id(self) -> Optional[str]:
        """
        Returns the Provider ID
        """
        if not self.is_authenticated: return None
        if self.method == 'apikey_user':
            return self.token.provider_id
        return self.token.identity
    
    @eproperty
    def email(self) -> Optional[str]:
        """
        Returns the User's Email
        """
        if not self.is_authenticated: return None
        if self.data and self.data.get('email'): return self.data.get('email')
        if self.token.email: return self.token.email
        if self.data:
            return self.data.get('preferred_email', self.data.get('email'))
        return None
    
    @email.setter
    def email(self, value: str):
        """
        Sets the User's Email
        """
        self._extra['email'] = value
        self.data['email'] = value

    @eproperty
    def email_domain(self) -> Optional[str]:
        """
        Returns the User's Email Domain
        """
        if not self.is_authenticated: return None
        if self.data and self.data.get('email_domain'): return self.data.get('email_domain')
        if hasattr(self.token, 'hd'): return self.token.hd
        if hasattr(self.token, 'email_domain'): return self.token.email_domain
        return self.email.split('@', 1)[-1] if self.email else None
    
    @email_domain.setter
    def email_domain(self, value: str):
        """
        Sets the User's Email Domain
        """
        self._extra['email_domain'] = value
        self.data['email_domain'] = value
        
    @eproperty
    def scopes(self) -> Optional[List[str]]:
        """
        Returns the User's Scopes
        """
        if not self.is_authenticated: return []
        if self.data and self.data.get('scopes'): return self.data.get('scopes')
        if self.token.scopes: return self.token.scopes
        return self.data.get('scopes') if self.data and self.data.get('scopes') else []
    
    @scopes.setter
    def scopes(self, value: List[str]):
        """
        Sets the User's Scopes
        """
        self._extra['scopes'] = value
        self.data['scopes'] = value

    def has_role(self, role: Union[str, int, UserRole]) -> bool:
        """
        Checks if the user has a role
        """
        if not self.is_authenticated: return False
        if not self.roles: return False
        if isinstance(self.roles[0], str): return role in self.roles
        if isinstance(self.roles[0], dict):
            return any(role in role['name'] for role in self.roles)
        if isinstance(self.roles[0], UserRole):
            return any(role >= role for role in self.roles)
        return False

    def add_role(self, role: Union[str, UserRole]):
        """
        Adds a role
        """
        if hasattr(self, f'is_{role}'): setattr(self, f'is_{role}', True)
        if isinstance(role, str): role = UserRole.parse_role(role)
        if not self.roles: self.roles = []
        if role not in self.roles: self.roles.append(role)

    def get_user_info(self) -> Dict[str, Any]:
        """
        Returns the User Info
        """
        return self.data

    def raise_if_not_authenticated(self):
        """
        Raises if the user is not authenticated
        """
        if not self.is_authenticated: 
            from .errors import NotAuthenticatedError
            raise NotAuthenticatedError()

    def __bool__(self) -> bool:
        """
        Returns True if the user is authenticated
        """
        return self.is_authenticated



class UnauthenticatedUser(UserMixin):
    """
    The Unauthenticated User
    """


class OAuth2User(UserMixin):
    """
    The OAuth2 User
    """
    _apikey: Optional[str] = PrivateAttr(None)
    
    @classmethod
    def initialize(
        cls,
        token: Optional['TokenT'] = None,
        provider: Optional['ProviderClientT'] = None,
        manager: Optional['OAuth2Manager'] = None,
        apikey: Optional[str] = None,
        **kwargs,
    ) -> 'OAuth2User':
        """
        Initializes the OAuth2 User
        """
        new = cls(
            token = token,
        )
        if provider: 
            new.client = provider
            new.provider = provider.name
        if manager: new.manager = manager
        new._apikey = apikey
        return new


    @eproperty
    def manager(self) -> 'OAuth2Manager':
        """
        Returns the OAuth2 Manager
        """
        return self._extra.get('manager')
    
    @eproperty
    def client(self) -> 'ProviderClientT':
        """
        Returns the Provider Client
        """
        return self._extra.get('client')

    async def postinit(self, request: Optional['Request'] = None, **kwargs):
        """
        Post Init
        """
        if not self.is_authenticated: return
        if self.manager.config.enable_user_info:
            await self.retrieve_user_info()
        if self.client is not None: 
            await self.client.set_user_data(self)
        await self.configure_perms()

    async def retrieve_user_info(self):
        """
        Retrieves the User Info
        """
        if self.client is None or self.client.config.user_info_retrieve_disabled: return
        # user_data = await self.client.afetch_user_info(self.provider_id)
        user_data = await self.client.afetch_user_info(self)
        if user_data: self.data = user_data

    async def configure_perms(self):
        """
        Configures the Permissions
        """
        if self.email:
            if self.email in self.manager.config.admin_emails:
                self.add_role('admin')
            elif self.email in self.manager.config.staff_emails:
                self.add_role('staff')
            elif self.email in self.manager.config.sa_emails:
                self.add_role('service_account')
            
        if self.email_domain:
            if self.email_domain in self.manager.config.admin_email_domains:
                self.add_role('admin')
            elif self.email_domain in self.manager.config.staff_email_domains:
                self.add_role('staff')
            elif self.email_domain in self.manager.config.sa_email_domains:
                self.add_role('service_account')
            

    def get_user_info(self) -> Dict[str, Any]:
        """
        Returns the User Info
        """
        data = {
            'method': self.method,
            'provider': self.provider,
            'provider_id': self.provider_id,
            'user': {
                'identity': self.identity,
                'email': self.email,
                'roles': self.roles,
                'scopes': self.scopes,
            },
            'role_types': {
                'is_admin': self.is_admin,
                'is_staff': self.is_staff,
                'is_service_account': self.is_service_account,
            },
        }
        if self.apikey: data['apikey'] = self.apikey
        if self.method == 'authentication': data['authorization'] = self.token.value
        if self.data: data['user'].update(self.data)
        return data
    
    @property
    def apikey(self) -> Optional[str]:
        """
        Returns the API Key
        """
        if 'apikey' not in self._extra:
            if self._apikey: self._extra['apikey'] = self._apikey
            else: self._extra['apikey'] = self.manager.encrypt_api_key(
                {
                    'identity': self.provider_id,
                    'provider': self.provider,
                }
            )
        return self._extra['apikey']


    @property
    def display_name(self) -> str:
        """
        Returns the User's Display Name
        """
        if not self.is_authenticated: return ""
        return self.data.get('name', self.data.get('given_name', '') + ' ' + self.data.get('family_name', ''))

    
    async def alogout(self):
        """
        Logs the User Out
        """
        if not self.is_authenticated: return

    @classmethod
    async def afrom_authorization(
        cls,
        token: 'TokenT',
        provider: 'ProviderClientT',
        manager: 'OAuth2Manager',
        request: 'Request',
        **kwargs,
    ) -> 'OAuth2User':
        """
        Creates the OAuth2 User from an Authorization
        """
        if token._token_type != 'oauth2_token':
            token = provider.config.oauth2_token_class(**{token._token_type: token})
            token.set_provider(provider.name)
        new = cls.initialize(token = token, provider = provider, manager = manager)
        new.method = 'authorization'
        await new.postinit(request = request)
        return new


    @classmethod
    async def afrom_session(
        cls,
        token: 'TokenT',
        provider: 'ProviderClientT',
        manager: 'OAuth2Manager',
        request: 'Request',
        **kwargs,
    ) -> 'OAuth2User':
        """
        Creates the OAuth2 User from a Session
        """
        new = cls.initialize(token = token, provider = provider, manager = manager)
        new.method = 'session'
        await new.postinit( request = request)
        return new
    
    @classmethod
    async def afrom_api_key_user(
        cls,
        api_key_user: 'APIKeyUser',
        provider: 'ProviderClientT',
        manager: 'OAuth2Manager',
        request: 'Request',
        apikey: str,
        **kwargs,
    ) -> 'OAuth2User':
        """
        Creates the OAuth2 User from an API Key User
        """
        new = cls.initialize(token = api_key_user, provider = provider, manager = manager, apikey = apikey)
        new.method = 'apikey_user'
        new.roles = [api_key_user.role] if api_key_user.role else []
        await new.postinit( request = request)
        return new
    

    @classmethod
    async def afrom_api_key(
        cls,
        token: 'TokenT',
        provider: 'ProviderClientT',
        manager: 'OAuth2Manager',
        request: 'Request',
        apikey: str,
        **kwargs,
    ) -> 'OAuth2User':
        """
        Creates the OAuth2 User from an API Key
        """
        new = cls.initialize(token = token, provider = provider, manager = manager, apikey = apikey)
        new.method = 'apikey'
        await new.postinit( request = request)
        return new

    def _refresh_from_state(
        self,
        provider: 'ProviderClientT',
        manager: 'OAuth2Manager',
        token: Optional['TokenT'] = None,
        apikey: Optional[str] = None,
        **kwargs,
    ):
        """
        Helper method to refresh the user from state
        """
        self.client = provider
        self.manager = manager
        self.method = 'state'
        if token is not None: self.token = token
        if apikey is not None: self._apikey = apikey


from __future__ import annotations

"""
FastAPI Helpers
"""

import functools
from lzl.proxied import ProxyObject
from typing import Optional, List, Dict, Any, Union, Type, Callable, TYPE_CHECKING

if TYPE_CHECKING:
    from lzl.io import JsonSerializer


json_serializer: 'JsonSerializer' = ProxyObject(
    obj_getter = 'lzl.io.ser.get_serializer',
    obj_args = ('json',),
    obj_kwargs = {'ser_mode': 'raw'},
)



from __future__ import annotations

"""
Schema Helpers for OpenAI
"""
import math
import functools
from lzl.types.base import get_pydantic_schema, PYDANTIC_VERSION
from typing import TYPE_CHECKING, Any, Sequence



if PYDANTIC_VERSION == 2:
    from pydantic.json_schema import GenerateJsonSchema as BaseGenerateJsonSchema

    if TYPE_CHECKING:
        from pydantic.config import JsonValue
        from pydantic_core import CoreSchema, PydanticOmit, core_schema
        from pydantic.json_schema import JsonSchemaValue, CoreSchemaField

    """
    Some type-specific keywords are not yet supported
    Notable keywords not supported include:

    For strings: minLength, maxLength, pattern, format
    For numbers: minimum, maximum, multipleOf
    For objects: patternProperties, unevaluatedProperties, propertyNames, minProperties, maxProperties
    For arrays: unevaluatedItems, contains, minContains, maxContains, minItems, maxItems, uniqueItems
    """

    hidden_fields = [
        'function_name',
        'function_model',
        'function_duration',
        'function_client_name',
    ]


    class GenerateJsonSchema(BaseGenerateJsonSchema):
        """
        Patched GenerateJsonSchema to support pydantic v2
        """
        _allow_examples: bool = True

        def int_schema(self, schema: 'core_schema.IntSchema') -> 'JsonSchemaValue':
            """Generates a JSON schema that matches an int value.

            Args:
                schema: The core schema.

            Returns:
                The generated JSON schema.
            """
            # currently not supported:
            # minimum, maximum, multipleOf
            json_schema: dict[str, Any] = {'type': 'integer'}
            self.update_with_validations(json_schema, schema, self.ValidationsMapping.numeric)
            json_schema = {k: v for k, v in json_schema.items() if v not in {math.inf, -math.inf}}
            for key in {'minimum', 'maximum', 'multipleOf'}:
                if key in json_schema:
                    _ = json_schema.pop(key)
            if not self._allow_examples:
                _ = json_schema.pop('examples', None)
            return json_schema
        

        def list_schema(self, schema: 'core_schema.ListSchema') -> 'JsonSchemaValue':
            """Returns a schema that matches a list schema.

            Args:
                schema: The core schema.

            Returns:
                The generated JSON schema.
            """
            # currently not supported:
            # unevaluatedItems, contains, minContains, maxContains, minItems, maxItems, uniqueItems
            items_schema = {} if 'items_schema' not in schema else self.generate_inner(schema['items_schema'])
            json_schema = {'type': 'array', 'items': items_schema}
            self.update_with_validations(json_schema, schema, self.ValidationsMapping.array)
            for key in {'unevaluatedItems', 'contains', 'minContains', 'maxContains', 'minItems', 'maxItems', 'uniqueItems'}:
                if key in json_schema:
                    _ = json_schema.pop(key)
            if not self._allow_examples:
                _ = json_schema.pop('examples', None)
            return json_schema
        

        def str_schema(self, schema: 'core_schema.StringSchema') -> 'JsonSchemaValue':
            """Generates a JSON schema that matches a string value.

            Args:
                schema: The core schema.

            Returns:
                The generated JSON schema.
            """
            # currently not supported:
            # minLength, maxLength, pattern, format
            json_schema = {'type': 'string'}
            self.update_with_validations(json_schema, schema, self.ValidationsMapping.string)
            for key in {'minLength', 'maxLength', 'pattern', 'format'}:
                if key in json_schema:
                    _ = json_schema.pop(key)
            if not self._allow_examples:
                _ = json_schema.pop('examples', None)
            return json_schema


        def default_schema(self, schema: 'core_schema.WithDefaultSchema'):
            """Generates a JSON schema that matches a schema with a default value.

            Args:
                schema: The core schema.

            Returns:
                The generated JSON schema.
            """
            # We remove the default value from the schema since it isn't needed
            # during the generation but can be used during validation.
            return self.generate_inner(schema['schema'])

        def handle_ref_overrides(self, json_schema: JsonSchemaValue) -> JsonSchemaValue:
            """It is not valid for a schema with a top-level $ref to have sibling keys.

            During our own schema generation, we treat sibling keys as overrides to the referenced schema,
            but this is not how the official JSON schema spec works.

            Because of this, we first remove any sibling keys that are redundant with the referenced schema, then if
            any remain, we transform the schema from a top-level '$ref' to use allOf to move the $ref out of the top level.
            (See bottom of https://swagger.io/docs/specification/using-ref/ for a reference about this behavior)
            """
            json_schema = super().handle_ref_overrides(json_schema)
            if 'allOf' in json_schema:
                ref = json_schema.pop('allOf')
                json_schema.update(ref[0])
                # Remove description from here
                if 'description' in json_schema:
                    _ = json_schema.pop('description')
            if 'title' in json_schema:
                _ = json_schema.pop('title')
            if not self._allow_examples:
                _ = json_schema.pop('examples', None)
            return json_schema

        def field_is_required(
            self,
            field: 'core_schema.ModelField' | 'core_schema.DataclassField' | 'core_schema.TypedDictField',
            total: bool,
        ) -> bool:
            """Whether the field should be marked as required in the generated JSON schema.
            (Note that this is irrelevant if the field is not present in the JSON schema.).

            Args:
                field: The schema for the field itself.
                total: Only applies to `TypedDictField`s.
                    Indicates if the `TypedDict` this field belongs to is total, in which case any fields that don't
                    explicitly specify `required=False` are required.

            Returns:
                `True` if the field should be marked as required in the generated JSON schema, `False` otherwise.
            """
            # if self.mode == 'serialization' and self._config.json_schema_serialization_defaults_required:
            #     return not field.get('serialization_exclude')
            # if field.
            # print(field)
            # print(field['metadata']['pydantic_js_annotation_functions'])
            return True

        def _named_required_fields_schema(
            self, named_required_fields: Sequence[tuple[str, bool, 'CoreSchemaField']],
        ):
            """
            Generate a JSON Schema for the required fields.
            """
            json_schema = super()._named_required_fields_schema(named_required_fields)
            json_schema['additionalProperties'] = False
            return json_schema
            

        # def model_fields_schema(self, schema: 'core_schema.ModelFieldsSchema') -> JsonSchemaValue:
        #     """Generates a JSON schema that matches a schema that defines a model's fields.

        #     Args:
        #         schema: The core schema.

        #     Returns:
        #         The generated JSON schema.
        #     """
        #     json_schema = super().model_fields_schema(schema)
        #     print(json_schema)
        #     if 'examples' in json_schema:
        #         _ = json_schema.pop('examples')
        #     return json_schema
            

    _excluded_fields = ['function_name', 'function_model', 'function_duration', 'function_client_name', 'function_usage']
    
    class FunctionGenerateJsonSchema(GenerateJsonSchema):
        _allow_examples: bool = False

        def _named_required_fields_schema(
            self, named_required_fields: Sequence[tuple[str, bool, 'CoreSchemaField']]
        ) -> JsonSchemaValue:
            """
            Generate a JSON Schema for the required fields.
            """
            required_fields = [
                (name, required, field)
                for name, required, field in named_required_fields
                if name not in _excluded_fields
            ]
            return super()._named_required_fields_schema(required_fields)

    
    generate_json_schema = functools.partial(get_pydantic_schema, schema_generator = GenerateJsonSchema)
    function_generate_json_schema = functools.partial(get_pydantic_schema, schema_generator = FunctionGenerateJsonSchema)
else:
    generate_json_schema = get_pydantic_schema
    function_generate_json_schema = get_pydantic_schema




from .config import SqliteConfig, SqliteSettings
from .backend import SqliteBackend, SQLAlchemyBackend, SQLModelBackend, SqliteBackendMap, SqliteBackendClasses
from .manager import SqliteBackendManager


"""Compatibility aliases for typing features across Python versions."""

import sys

if sys.version_info >= (3, 8):
    from typing import Final, Literal
else:  # pragma: no cover - legacy Python support
    from typing_extensions import Final, Literal

__all__ = ['Final', 'Literal']
